####
####
##
##
## Copyright (C) 2008 Andreas Heger All rights reserved
##
## Author: Andreas Heger <andreas.heger@helsinki.fi>
##
## $Id: psl2psl.py 2781 2009-09-10 11:33:14Z andreas $
##
##
####
####
'''
psl2psl.py - manipulate psl files
===================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script reads alignments in :term:`psl` format,
applies one ore more operations on them and then
outputs them again in :term:`psl`.

The following operators are available:

merge
   merges psl alignments between overlapping entries of query/target
        in a file. Merging is stopped on query/target and/or strand change.
        This script will check for overlaps and will take the longest path
        through a set of overlapping paths.

map 
   map alignments. Similar to mapPsl, but will apply a map to both query
   and target. This method requires the two options filter-query and filter-target.
   Intervals are given in gff or gtf format. In the latter case, the alignment is 
   threaded through the exons. 

filter-keep
   filter alignments. Only parts of the alignments are kept that are part of the intervals.

filter-remove
   filter alignments. Only parts of the alignments are kept that are not part of the intervals.

complement
   complement a psl sequence. For example this will convert exons to introns.
   If an entry contains only a single exon, it is omitted.

select-longest
   select only the longest transcript (nmatches)

rename-query
   rename matches (query) consecutively.

test
   go through file and remove all non-parseable lines

sanitize
   change query/target names according to fasta sequence database

remove-overlapping-query
   remove all alignments that are not unique with respect to query segment
   (requires alignments to be sorted by query)

remove-overlapping-target
   remove all alignments that are not unique with respect to a target segment
   (requires alignments to be sorted by target)

Usage
-----

Example::

   python <script_name>.py --help

Type::

   python <script_name>.py --help

for command line help.

Documentation
-------------

Code
----
'''


import sys
import re
import string
import optparse
import time
import os
import collections
import warnings

import CGAT.Experiment as E
import CGAT.Stats as Stats
import CGAT.Genomics as Genomics
import CGAT.Blat as Blat
import CGAT.GTF as GTF
import CGAT.IndexedFasta as IndexedFasta
import CGAT.IndexedGenome as IndexedGenome

## ignore warnings from networkx/matplotlib that a display
## can not be found
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import networkx

import alignlib
import bx.intervals.io
import bx.intervals.intersection

def readIntervals( infile, options ):

    ninput = 0
    t = time.time()

    if options.format == "gtf":

        index = IndexedGenome.IndexedGenome()

        for gffs in GTF.transcript_iterator( GTF.iterator(infile) ):
        
            ali = alignlib.makeAlignmentBlocks()
            for gff in gffs:
                if gff.feature != "exon": continue
                ali.addDiagonal( gff.start, gff.end, 0 )

            index.add( min( [x.start for x in gffs]), 
                       max( [x.end for x in gffs] ),
                       ali )
            ninput += 1
        
            if ninput % options.report_step == 0:
                E.info("reading intervals - progress: ninput=%i, time=%i, avg=%f" % (ninput, time.time() -t, float(time.time() -t) / ninput))
                
    elif options.format == "gff":

        index = IndexedGenome.Simple()

        for g in GTF.iterator(infile):
        
            index.add( g.contig, g.start, g.end )
            ninput += 1

            if ninput % options.report_step == 0:
                E.info("reading intervals - progress: ninput=%i, time=%i, avg=%f" % (ninput, time.time() -t, float(time.time() -t) / ninput))

    E.info("read intervals: %i contigs, %i intervals" % (len(index), ninput))
    return index

def iterator_psl_intervals( options ):
    """iterate over psl file yield an entry together with overlapping entries.

    returns tuples of (match, list(query_intervals), list(target_intervals))
    """

    if options.filename_filter_query:
        intervals_query = readIntervals( open(options.filename_filter_query, "r" ), options )
    else:
        intervals_query = None

    if options.filename_filter_target:
        intervals_target = readIntervals( open(options.filename_filter_target, "r" ), options )
    else:
        intervals_target = None
        
    iterator = Blat.BlatIterator( options.stdin )

    ninput = 0

    while 1:
        
        match = iterator.next()
        if not match: break

        ninput += 1        
        if options.test and ninput >= options.test:
            break

        if options.loglevel >= 1 and ninput % options.report_step == 0:
            options.stdlog.write( "# progress: ninput=%i\n" % (ninput))
            options.stdlog.flush()

        qx, tx = None, None
        if intervals_query:
            try:
                qx = list(intervals_query.get( match.mQueryId, match.mQueryFrom, match.mQueryTo ))
            except KeyError:
                qx = []

        if intervals_target:
            try:
                tx = list(intervals_target.get( match.mSbjctId, match.mSbjctFrom, match.mSbjctTo ))
            except KeyError:
                tx = []

        if options.loglevel >= 2:
            options.stdlog.write( "###################################################\n" )
            options.stdlog.write( "# testing match %s\n" % (str(match)))
            options.stdlog.write( "###################################################\n" )

        yield match, qx, tx

def pslFilter( options, keep = True ):
    """filter psl entries. 

    Only positions contained in intervals are kept, unless remove
    is set, in which case only positions not in intervals are kept.

    """

    ## read as gff
    options.format = "gff"

    ninput, noutput, ndiscarded, nskipped = 0, 0, 0, 0

    if keep:
        raise "not implemented"
    else:
        for match, qx, tx in iterator_psl_intervals( options ):

            map_query2target = match.getMapQuery2Target()

            if not qx and not tx:
                options.stdout.write( str(match) + "\n" )
                noutput += 1
                continue
            
            if qx:
                for qstart, qend, v in qx:
                    if match.strand == "-":
                        qstart, qend = match.mQueryLength - qend, match.mQueryLength - qstart
                    map_query2target.removeRowRegion( qstart, qend )
        
            if tx:
                for tstart, tend, v in tx:
                    map_query2target.removeColRegion( tstart, tend )
            
            if map_query2target.getNumAligned() > options.min_aligned:
                match.fromMap( map_query2target, use_strand = True )
                options.stdout.write( str(match) + "\n" )
                noutput += 1
            else:
                ndiscarded += 1

    E.info("ninput=%i, noutput=%i, nskipped=%i, ndiscarded=%i" % (ninput, noutput, nskipped, ndiscarded) )

def pslMap( options ):
    """thread psl alignments using intervals.

    """

    if options.format == "gtf":
        use_copy = False
    else:
        use_copy = True

    ninput, noutput, ndiscarded, nskipped, nskipped_small_queries = 0, 0, 0, 0, 0

    min_length = options.min_aligned

    for match, qx, tx in iterator_psl_intervals( options ):

        map_query2target = match.getMapQuery2Target()

        ninput += 1

        ## if no filter on qx or tx, use full segment
        if qx == None:
            qx = [ (match.mQueryFrom,match.mQueryTo,0) ]
        elif tx == None:
            tx = [ (match.mSbjctFrom,match.mSbjctTo,0) ]

        ## if no overlap: return
        if not qx or not tx: 
            nskipped += 1
            continue

        for query in qx:

            qstart, qend, qval = query

            # skip elements that are too small
            if qend - qstart < min_length: 
                E.debug( "query too small - skipped at %s:%i-%i" % (match.mQueryId, qstart, qend) )
                nskipped_small_queries += 1
                continue

            E.debug( "working on query %s:%i-%i" % (match.mQueryId, qstart, qend) )

            mqstart, mqend = ( map_query2target.mapRowToCol(qstart, 
                                                            alignlib.RIGHT), 
                               map_query2target.mapRowToCol(qend, 
                                                            alignlib.LEFT) )
                        
                
            if match.strand == "-":
                qstart, qend = match.mQueryLength - qend, match.mQueryLength - qstart

            for target in tx:

                tstart, tend, tval = target
                if tstart >= mqend or tend <= mqstart: continue
                if tend - tstart < min_length: continue

                new = alignlib.makeAlignmentBlocks()
                    
                if use_copy:
                    # do copy with range filter
                    if options.loglevel >= 3:

                        mtstart, mtend = map_query2target.mapColToRow(tstart), map_query2target.mapColToRow(tend) 
                        E.debug( "query: %i-%i (len=%i)-> %i-%i(len=%i); target: %i-%i (len=%i)-> %i-%i (len=%i)" % \
                                     (qstart, qend,
                                      qend - qstart,
                                      mqstart, mqend,
                                      mqend - mqstart,
                                      tstart, tend,
                                      tend - tstart,
                                      mtstart, mtend,
                                      mtend - mtstart ) )
                                     
                    alignlib.copyAlignment( 
                        new, 
                        map_query2target,
                        qstart, qend,
                        tstart, tend )
                else:
                    # do copy with alignment filter
                    map_query = qval
                    if map_query:
                        tmp = alignlib.makeAlignmentBlocks()                        
                        alignlib.copyAlignment( tmp, map_query2target, map_query, alignlib.RR )
                        if options.loglevel >= 5:
                            options.stdlog.write( "######## mapping query ###########\n" )
                            options.stdlog.write( "# %s\n" % str(alignlib.AlignmentFormatEmissions( map_query2target ) ))
                            options.stdlog.write( "# %s\n" % str(alignlib.AlignmentFormatEmissions( map_query ) ))
                            options.stdlog.write( "# %s\n" % str(alignlib.AlignmentFormatEmissions( tmp ) ))
                    else:
                        tmp = map_query2target
                        
                    map_target = tval
                    if map_target:
                        new = alignlib.makeAlignmentBlocks()
                        alignlib.copyAlignment( new, tmp, map_target, alignlib.CR )                        
                        if options.loglevel >= 5:
                            options.stdlog.write( "######## mapping target ###########\n" )
                            options.stdlog.write( "# before: %s\n" % str(alignlib.AlignmentFormatEmissions( tmp ) ))
                            options.stdlog.write( "# map   : %s\n" % str(alignlib.AlignmentFormatEmissions( map_target ) ))
                            options.stdlog.write( "# after : %s\n" % str(alignlib.AlignmentFormatEmissions( new ) ))
                    else:
                        new = tmp

                if options.loglevel >= 4:
                    E.debug("putative match with intervals: %s and %s: %i-%i" % \
                                (str(query), str(target), qstart, qend ))
                    if options.loglevel >= 5:
                        E.debug( "input : %s" % str(alignlib.AlignmentFormatEmissions( map_query2target ) ))
                        E.debug( "final : %s" % str(alignlib.AlignmentFormatEmissions( new ) ) )

                    if new.getLength() > 0:
                        n = match.copy()
                        n.fromMap( new, use_strand = True )
                        E.info( "match : %s" % (str(n)))

                if new.getNumAligned() > options.min_aligned:
                    n = match.copy()
                    n.fromMap( new, use_strand = True )
                    options.stdout.write( str(n) + "\n" )
                    noutput += 1
                else:
                    ndiscarded += 1

    E.info( "map: ninput=%i, noutput=%i, nskipped=%i, ndiscarded=%i, nsmall_queries=%i" % \
                (ninput, noutput, nskipped, ndiscarded, nskipped_small_queries) )

def pslMerge( options ):
    """merge psl alignments.
    """

    iterator = Blat.BlatIterator( sys.stdin )

    ninput, noutput, ndiscarded, nskipped = 0, 0, 0, 0

    last_query = None
    last_target = None
    last_strand = None

    def process( matches ):

        new = matches[0].copy()

        map_query2target = alignlib.makeAlignmentBlocks()

        graph = networkx.DiGraph()
        graph.add_nodes_from(xrange(len(matches)+2))

        matches.sort( key = lambda x: x.mQueryFrom )

        if Genomics.IsPositiveStrand( matches[0].strand ):
            f = lambda x,y: x.mSbjctTo < y.mSbjctFrom
        else:
            f = lambda x,y: x.mSbjctFrom > y.mSbjctTo

        for x in range(0,len(matches)):

            xx = matches[x]
            if options.loglevel >= 6:
                options.stdlog.write( "# graph: %2i %s\n" % (x, str(xx)))

            for y in range(x+1,len(matches)):
                yy = matches[y]
                d = min(xx.mQueryTo,yy.mQueryTo) - max(xx.mQueryFrom,yy.mQueryFrom)
                if d > 0 or not f(xx,yy):
                    continue
                else:
                    graph.add_edge( x, y, { 'weight' : -d } )

        source = len(matches)
        target = len(matches) + 1
        for x in range(len(matches)):
            xx = matches[x]
            graph.add_edge( source, x, { 'weight': xx.mQueryFrom } ) 
            graph.add_edge( x, target, { 'weight' : xx.mQueryLength - xx.mQueryTo } )

        if options.loglevel >= 6:
            networkx.write_edgelist( graph, options.stdlog )

        path = networkx.dijkstra_path( graph, source, target ) 

        if options.loglevel >= 6:
            options.stdlog.write( "# path: %s\n" % (str(path)))

        new_matches = [ matches[x] for x in path[1:-1] ]

        if len(matches) != len(new_matches):
            E.warn( "query=%s, target=%s, strand=%s: removed overlapping/out-of-order segments: before=%i, after=%i" %\
                                          (matches[0].mQueryId,
                                           matches[0].mSbjctId,
                                           matches[0].strand,
                                           len(matches),
                                           len(new_matches) ) )

        matches = new_matches

        for match in matches:
            m = match.getMapQuery2Target()
            alignlib.addAlignment2Alignment( map_query2target, m )

        new.fromMap( map_query2target, use_strand = True )
        
        options.stdout.write( str( new ) + "\n" )
        options.stdout.flush()
        return 1

    while 1:
        
        match = iterator.next()
        if not match: break

        ninput += 1        
        if options.test and ninput >= options.test:
            break

        if options.loglevel >= 10:
            options.stdlog.write( "# input: %s\n" % (str(match)))

        if ninput % options.report_step == 0:
            E.info("progress: ninput=%i, noutput=%i" % (ninput, noutput))

        if match.mQueryId != last_query or match.strand != last_strand or match.mSbjctId != last_target:
            if last_query:
                noutput += process( matches )
            matches = []
            last_query, last_target, last_strand = match.mQueryId, match.mSbjctId, match.strand
        
        matches.append( match )
            
    if last_query:
        noutput += process( matches )

    E.info("ninput=%i, noutput=%i, nskipped=%i, ndiscarded=%i" % (ninput, noutput, nskipped, ndiscarded) )

def pslAddSequence( query_fasta, sbjct_fasta, options ):

    iterator = Blat.BlatIterator( sys.stdin )

    ninput, noutput, ndiscarded, nskipped = 0, 0, 0, 0


    while 1:
        
        match = iterator.next()
        if not match: break

        ninput += 1        
        if options.test and ninput >= options.test:
            break

        if ninput % options.report_step == 0:
            E.info("progress: ninput=%i, noutput=%i" % (ninput, noutput))

        new = Blat.MatchPSLX()
        new.fromPSL( match,
                     query_fasta.getSequence( match.mQueryId, "+", match.mQueryFrom, match.mQueryTo ),
                     sbjct_fasta.getSequence( match.mSbjctId, "+", match.mSbjctFrom, match.mSbjctTo ) )
        
        options.stdout.write( str( new ) + "\n" )
        noutput += 1

    E.info("ninput=%i, noutput=%i, nskipped=%i, ndiscarded=%i" % (ninput, noutput, nskipped, ndiscarded) )

def pslComplement( query_fasta, target_fasta, options) :
    """complenment psl entries.
    """

    iterator = Blat.BlatIterator( sys.stdin )

    ninput, noutput, ndiscarded, nskipped = 0, 0, 0, 0

    border = options.complement_border
    min_length = options.complement_min_length

    while 1:
        
        match = iterator.next()
        if not match: break

        ninput += 1        
        if options.test and ninput >= options.test:
            break

        if ninput % options.report_step == 0:
            E.info("progress: ninput=%i, noutput=%i" % (ninput, noutput))

        if match.mNBlocks <= 1: 
            nskipped += 1
            continue

        pairs = []
        for qstart, tstart, size in match.getBlocks():

            qend = qstart + size - border
            qstart += border
            
            if qend - qstart < options.complement_min_length:
                continue

            tend = tstart + size - border
            tstart += border

            if tend - tstart < options.complement_min_length:
                continue

            query_sequence = query_fasta.getSequence( match.mQueryId, match.strand, qstart, qend )
            sbjct_sequence = sbjct_fasta.getSequence( match.mSbjctId, "+", tstart, tend )
            

        ndiscarded += 1


        options.stdout.write( str( new ) + "\n" )
        noutput += 1

    E.info("ninput=%i, noutput=%i, nskipped=%i, ndiscarded=%i" % (ninput, noutput, nskipped, ndiscarded) )

def pslComplementQuery( options) :
    """complement psl entries. 

    Fill the regions from a second psl file. 
    """

    iterator = Blat.BlatIterator( sys.stdin )

    ninput, noutput, ndiscarded, nskipped = 0, 0, 0, 0

    border = options.complement_border
    min_length = options.complement_min_length

    while 1:
        
        match = iterator.next()
        if not match: break

        ninput += 1        
        if options.test and ninput >= options.test:
            break

        if ninput % options.report_step == 0:
            E.info("progress: ninput=%i, noutput=%i" % (ninput, noutput))

        if match.mNBlocks <= 1: 
            nskipped += 1
            continue

        pairs = []
        for qstart, tstart, size in match.getBlocks():

            qend = qstart + size - border
            qstart += border
            
            if qend - qstart < options.complement_min_length:
                continue

            tend = tstart + size - border
            tstart += border

            if tend - tstart < options.complement_min_length:
                continue

        ndiscarded += 1


        options.stdout.write( str( new ) + "\n" )
        noutput += 1

    E.info("ninput=%i, noutput=%i, nskipped=%i, ndiscarded=%i" % (ninput, noutput, nskipped, ndiscarded) )

def pslSelectQuery( options ):

    ninput, noutput, ndiscarded, nskipped = 0, 0, 0, 0

    value, field = options.select.split( "-" )

    if field == "nmatches":
        f = lambda x: x.mNMatches
    elif field == "nmismatches":
        f = lambda x: x.mNMisMatches

    for data in Blat.iterator_per_query( Blat.iterator( options.stdin ) ):
        
        ninput += 1        
        if options.test and ninput >= options.test:
            break

        if ninput % options.report_step == 0:
            E.info("progress: ninput=%i, noutput=%i" % (ninput, noutput))

        data.sort( key = f )

        if value == "most":
            options.stdout.write( "%s\n" % str(data[-1]) )
        elif value == "least":
            options.stdout.write( "%s\n" % str(data[0]) )
            
        noutput += 1

    E.info("ninput=%i, noutput=%i, nskipped=%i, ndiscarded=%i" % (ninput, noutput, nskipped, ndiscarded) )

def iterator_filter_overlapping_query( psls, options ):
    '''remove alignments that overlap on query.

    If multiple alignments overlap, the one with the highest number
    of matching nucleotides is chosen.
    '''

    # note: only takes the full ranges, but does not check for individual overlap of blocks
    # use connected components and hasAlignmentOverlap
    ninput, noutput, ndiscarded = 0, 0, 0    

    last_contig = None

    for block in Blat.iterator_query_overlap( psls, options.threshold_merge_distance ):

        # commented code is for base-level filtering, which is very slow
        # disabled for now
        # if block[0].mQueryId != last_contig:
        #     last_contig = block[0].mQueryId
        #     E.info( "processing %s" % last_contig )

        l = len(block)
        ninput += l
        if l > 1: 
            ndiscarded += l
            # components = Blat.getComponents( block, by_query = True )
            # for component in components:
            #     m = [ block[x] for x in component ]
            #     m.sort( key = lambda x: -x.mNMatches )
            #     ndiscarded += len(m) - 1
            #     yield m[0]
            #     noutput += 1
        else:
            yield block[0]
            noutput += 1

    E.info( "iterator_filter_overlapping_query: ninput=%i, noutput=%i, ndiscarded=%i" % (ninput, noutput,ndiscarded) )

def iterator_filter_overlapping_target( psls, options ):

    ninput, noutput, ndiscarded = 0, 0, 0    
    for block in Blat.iterator_target_overlap( psls, options.threshold_merge_distance ):
        l = len(block)
        ninput += l
        if l > 1: 
            ndiscarded += l
        else:
            yield block[0]
            noutput += 1

    E.info( "iterator_filter_overlapping_target: ninput=%i, noutput=%i, ndiscarded=%i" % (ninput, noutput,ndiscarded) )

def iterator_rename_query( infile, options ):

    ninput, noutput, nerrors = 0, 0, 0

    map_old2new = {}
    x = 1
    while 1:

        match = infile.next()

        if not match: break
        ninput += 1

        if match.mQueryId not in map_old2new or options.unique:
            new = options.id_format % x
            map_old2new[ match.mQueryId ] = new
            x += 1
        else:
            new = map_old2new[match.mQueryId]
            match.mQueryId = new
        yield match

    if options.output_filename_map:
        outfile = open(options.output_filename_map, "w")
        outfile.write( "%s\t%s\n" % ("old", "new") )
        for old,new in map_old2new.iteritems():
            outfile.write( "%s\t%s\n" % (old, new) )
        outfile.close()

    E.info( "ninput=%i, noutput=%i, nerrors=%i" % (ninput, noutput,nerrors) )

def iterator_sanitize( infile, query_fasta, sbjct_fasta, options ):

    fq = query_fasta.getToken
    ft = sbjct_fasta.getToken

    ninput, noutput, nerrors = 0, 0, 0

    while 1:
        try:
            x = infile.next()
        except Blat.ParsingError, msg:
            nerrors += 1
            ninput += 1
            E.warn( str(msg) )
            continue
        except StopIteration:
            break

        if not x: break

        ninput += 1        

        if query_fasta: x.mQueryId = fq( x.mQueryId )
        if sbjct_fasta: x.mSbjctId = ft( x.mSbjctId )

        if ninput % options.report_step == 0:
            E.info("progress: ninput=%i, noutput=%i" % (ninput, noutput))

        yield x

    E.info( "ninput=%i, noutput=%i, nerrors=%i" % (ninput, noutput,nerrors) )

def iterator_filter_fasta( infile, query_fasta, sbjct_fasta, options ):
    
    ninput, noutput, nerrors = 0, 0, 0
    
    qmissing, smissing = collections.defaultdict(int), collections.defaultdict(int),
    
    while 1:
        try:
            x = infile.next()
        except Blat.ParsingError, msg:
            nerrors += 1
            ninput += 1
            E.warn( str(msg) )
            continue
        except StopIteration:
                    break

        if not x: break

        ninput += 1        
        skip = False
        if query_fasta: 
            if x.mQueryId not in query_fasta:
                qmissing[x.mQueryId] += 1
                skip = True
        if sbjct_fasta: 
            if x.mSbjctId not in sbjct_fasta:
                smissing[x.mSbjctId] += 1
                skip = True
        if skip:
            nerrors += 1
            continue

        if ninput % options.report_step == 0:
            E.info("progress: ninput=%i, noutput=%i" % (ninput, noutput))

        yield x
        noutput += 1

    if qmissing:
        E.warn( "query: %i ids missing: %s" % (len(qmissing), str(qmissing)))
    if smissing:
        E.warn( "target: %i ids missing: %s" % (len(smissing), str(smissing)))

    E.info( "ninput=%i, noutput=%i, nerrors=%i" % (ninput, noutput,nerrors) )

if __name__ == '__main__':

    parser = E.OptionParser( version = "%prog version: $Id: psl2psl.py 2781 2009-09-10 11:33:14Z andreas $", usage = globals()["__doc__"] )

    parser.add_option("--filter-query", dest="filename_filter_query", type="string",
                      help="filename with intervals in the query to filter (in gff format) [default=%default]."  )

    parser.add_option("--filter-target", dest="filename_filter_target", type="string",
                      help="filename with intervals in the target to filter (in gff format) [default=%default]."  )


    parser.add_option("-m", "--method", dest="methods", type="choice", action="append",
                      choices=("map", "merge", 
                               "add-sequence", "complement", 
                               "select-query", "test", 
                               "filter-keep", "filter-remove", 
                               "rename-query",
                               "sanitize",
                               "filter-fasta",
                               "remove-overlapping-query",
                               "remove-overlapping-target"),
                      help="""action to perform [default=%default].""" )

    parser.add_option("--select", dest="select", type="choice",
                      choices=("most-nmatches", "least-nmatches", "most-nmismatches", "least-nmismatches" ),
                      help="entry to select [default=%default]."  )

    parser.add_option("--header", dest="header", type="choice",
                      choices=( "none", "table", "full" ),
                      help="output psl header [default=%default]."  )

    parser.add_option("--format", dest="format", type="choice",
                      choices=("gff", "gtf" ),
                      help="format of intervals [default=%default]."  )

    parser.add_option("--filename-queries", dest="filename_queries", type="string",
                      help="fasta filename with queries."  )

    parser.add_option("--filename-target", dest="filename_sbjcts", type="string",
                      help="fasta filename with sbjct [default=%default]."  )

    parser.add_option("--id-format", dest="id_format", type="string",
                      help="format of new identifiers for the rename function [default=%default]."  )

    parser.add_option("--unique", dest="unique", action="store_true",
                      help="in the rename function, make each match unique [default=%default]."  )

    parser.add_option("--output-filename-map", dest="output_filename_map", type="string",
                      help="filename with map of old to new labels for rename function [default=%default]."  )

    parser.add_option("--complement-min-length", dest="complement_min_length", type="int",
                      help="minimum length for complemented blocks [default=%default]."  )

    parser.add_option("--complement-border", dest="complement_border", type="int",
                      help="number of residues to exclude before alignment at either end [default=%default]."  )

    parser.add_option("--complement-aligner", dest="complement_aligner", type="choice",
                      choices=("clustal", "dba", "dialign", "dialign-lgs" ),
                      help="aligner for complemented segments [default=%default]."  )

    parser.add_option( "--threshold-merge-distance", dest="threshold_merge_distance", type="int",
                       help="distance in nucleotides at which two adjacent reads shall be merged even if they are not overlapping [%default]." )

    parser.set_defaults( filename_filter_target = None,
                         filename_filter_query = None,
                         filename_queries = None,
                         filename_sbjcts = None,
                         threshold_merge_distance = 0,
                         report_step = 100000,
                         min_aligned = 100,
                         methods = [],
                         format = "gff",
                         select = "most-nmatches",
                         id_format = "%06i",
                         unique = False,
                         output_filename_map = None,
                         header = None,
                         test = None )
    
    (options, args) = E.Start( parser, add_pipe_options = True )

    if options.filename_queries:
        query_fasta =  IndexedFasta.IndexedFasta( options.filename_queries )
    else:
        query_fasta = None

    if options.filename_sbjcts:
        sbjct_fasta =  IndexedFasta.IndexedFasta( options.filename_sbjcts )
    else:
        sbjct_fasta = None

    if "add-sequence" in options.methods and (sbjct_fasta == None or query_fasta == None):
        raise ValueError( "please supply both indexed query and target/genome sequence data." )

    iterator = Blat.iterator( options.stdin )

    if options.header != None or options.header != "none":
        if options.header == "table":
            options.stdout.write( "\t".join( Blat.FIELDS ) + "\n" )
        elif options.header == "full":
            options.stdout.write( Blat.HEADER + "\n" )

    for method in options.methods:
    
        if "map" == method:
            pslMap( options )
            break
        elif "filter-keep" == method:
            pslFilter( options, keep = True )
            break
        elif "filter-remove" == method:
            pslFilter( options, keep = False )
            break
        elif "merge" == method:
            pslMerge( options )
            break
        elif "add-sequence" == method:
            pslAddSequence( query_fasta, sbjct_fasta, options )
            break
        elif "complement" == method:
            pslComplement( query_fasta, sbjct_fasta, options )
            break
        elif "select-query" == method:
            pslSelectQuery( options )
            break
        elif "test" == method:
            iterator = Blat.iterator_test( iterator, options.report_step )
        elif "rename-query" == method:
            iterator = iterator_rename_query( iterator, options )
        elif "sanitize" == method:
            iterator = iterator_sanitize( iterator, query_fasta, sbjct_fasta, options )
        elif "filter-fasta" == method:
            iterator = iterator_filter_fasta( iterator, query_fasta, sbjct_fasta, options )
        elif "remove-overlapping-query" == method:
            iterator = iterator_filter_overlapping_query( iterator, options )
        elif "remove-overlapping-target" == method:
            iterator = iterator_filter_overlapping_target( iterator, options )

    for psl in iterator:
        options.stdout.write( "%s\n" % str( psl ) )

    E.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
mali_phylip2fasta.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python mali_phylip2fasta.py --help

Type::

   python mali_phylip2fasta.py --help

for command line help.

Documentation
-------------

Code
----

'''
import sys
import string
import re

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $", 
                                    usage = globals()["__doc__"] )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv )

    line = sys.stdin.readline()
    num_lines, width = map(int, re.search("(\d+)\s+(\d+)", line[:-1]).groups())

    alignment = []

    for x in range(num_lines):
        line = sys.stdin.readline()
        id, ali = re.search( "^(\S+)\s+(.+)", line[:-1]).groups()
        alignment.append( (id, [ali]) )

    while 1:
        line = sys.stdin.readline()
        if not line: break
        for x in range(num_lines):
            line = sys.stdin.readline()            
            alignment[x][1].append( line[:-1] )

    for x in range(num_lines):
        print ">%s\n%s" % (alignment[x][0], re.sub("\s", "", string.join(alignment[x][1], "")))

    ## write footer and output benchmark information.
    E.Stop()




################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
histogram2histogram.py - operations on histograms
=================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script reads a histogram from stdin and
outputs a modified histogram on stdout. Modifications
are normalization, the cumulative, etc.

Usage
-----

Example::

   python histogram2histogram.py --help

Type::

   python histogram2histogram.py --help

for command line help.

Documentation
-------------

Code
----

'''
import sys
import re
import string
import os
import optparse

USAGE = """python %s < stdin > stdout

read in data and append columns to a density histogram

-> relative frequencies
-> cumulative counts and frequencies in both directions

'#' at start of line is a comment

""" % sys.argv[0]

import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import numpy

if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: histogram2histogram.py 2782 2009-09-10 11:40:29Z andreas $")

    parser.add_option("-i", "--is-int", dest="is_ints", action="store_true",
                      help="categories are integers."  )
    parser.add_option("-m", "--method", dest="method", type="choice",
                      choices = ("append", "cumul", "rcumul", "normalize" ),
                      help="method(s) to apply." )
    parser.add_option("--no-headers", dest="headers", action="store_false",
                      help="histogram has no headers.")
    parser.add_option("-c", "--columns", dest="columns", type="string",
                      help="columns to use for plotting.")
    parser.add_option("", "--truncate", dest="truncate", type="string",
                      help="truncate at range."  )
    parser.add_option("", "--no-out-of-range", dest="cumulate_out_of_range", action = "store_false",
                      help="add up bins out of range."  )
    parser.add_option( "--format-bin", dest="format_bin", type="string",
                      help="format for bins."  )
    parser.add_option( "--format-val", dest="format_val", type="string",
                      help="format for vals."  )

    parser.set_defaults(
        is_ints = False,
        method = "append",
        columns = "all",
        headers = True,
        truncate = None,
        cumulate_out_of_range = True,
        format_bin="%6.4f",
        format_val="%6.4f",
        )

    (options, args) = E.Start( parser )

    # old histogram2histogram.py semantics - need to merged with newer
    # code below.
    if options.method == "append":

        vals = []

        # retrieve histogram
        lines = filter( lambda x: x[0] <> "#", sys.stdin.readlines())

        # check if first line contains a header
        d = string.split(lines[0][:-1], "\t")[0]
        try:
            if options.is_ints:
                value = int(d)
            else:
                value = float( d )
        except ValueError:
            print string.join( (d, "counts", "frequency",
                                "cumulative counts", "increasing cumulative frequency",
                                "cumulative counts", "decreasing cumulative frequency"), "\t")
            del lines[0]

        data = map( lambda x: map(float, string.split(x[:-1], "\t")), lines)

        if len(data) == 0:
            raise "No data found."

        total = float(reduce( lambda x,y: x+y, map( lambda x: x[1], data)))

        cumul_down = int(total)
        cumul_up = 0

        if options.is_ints:
            form = "%i\t%i\t%6.4f\t%i\t%6.4f\t%i\t%6.4f"         
        else:
            form = "%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f" 

        for bin,val in data:
            percent     = float(val) / total
            cumul_up   += val
            percent_cumul_up = float(cumul_up) / total
            percent_cumul_down = float(cumul_down) / total        

            print form % \
                  (bin, val, percent, cumul_up, percent_cumul_up, cumul_down, percent_cumul_down)

            cumul_down -= val
        
    else:

        if options.truncate: options.truncate = map(float, options.truncate.split(","))

        options.method = options.method.split(",")
        data, legend = IOTools.readTable( sys.stdin,
                                          numeric_type=numpy.float32,
                                          take=options.columns,
                                          headers = options.headers,
                                          truncate= options.truncate,
                                          cumulate_out_of_range = options.cumulate_out_of_range )

        nfields = len(legend)

        ## note: because of MA, iteration makes copy of slices
        ## Solution: inplace edits.
        nrows, ncols = data.shape

        for method in options.method:
            if method == "cumul":
                l = [0] * ncols
                for x in range(nrows):
                    for y in range(1, ncols):
                        data[x,y] += l[y]
                        l[y] = data[x,y]

            elif method == "rcumul":
                l = [0] * ncols
                for x in range(nrows-1,0,-1):
                    for y in range(1, ncols):
                        data[x,y] += l[y]
                        l[y] = data[x,y]

            elif method == "normalize":
                m = [0] * ncols
                for x in range(nrows):
                    for y in range(1, ncols):
                        ## the conversion to float is necessary
                        m[y] = max( m[y], float(data[x,y]) )

                for y in range(1, ncols):
                    if m[y] == 0: m[y] = 1.0

                for x in range(nrows):
                    for y in range(1, ncols):
                        data[x,y] = data[x,y] / m[y]
            else:
                raise "unknown method %s" % method

        print "\t".join(legend)

        format = options.format_bin + "\t" + "\t".join( [options.format_val] * (nfields-1)) 

        for d in data:
            print format % tuple(d)
    
    E.Stop()












################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
cgat_script_template.py
=============================================

:Author: 
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

Usage
-----

Example::

   python cgat_script_template.py --help

Type::

   python cgat_script_template.py --help

for command line help.

Documentation
-------------

Code
----

'''

import os
import sys
import re
import optparse
import urllib2
import urllib
import collections
import xml.etree.ElementTree as ET

import CGAT.Experiment as E
import CGAT.IOTools as IOTools


def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $", 
                                    usage = globals()["__doc__"] )

    parser.add_option("-t", "--test", dest="test", type="string",
                      help="supply help"  )


    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv )

    params = { 'db' : 'gds',
               'term' : 'transcriptome[All Fields] AND "Homo sapiens"[Organism] AND high throughput sequencing[Platform Technology Type]',
               'retmax' : 5000,
               'usehistory' : 'y',
               }

    params = urllib.urlencode( params )
    query_filter = 'http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi'
    query_retrieve = 'http://eutils.ncbi.nlm.nih.gov/entrez/eutils/elink.fcgi'
    query_fetch = 'http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi'
    query_summary = 'http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi'
    
    data = urllib2.urlopen( query_filter, params )
    etree = ET.parse(data)
    root = etree.getroot()
    
    webenv = root.find( "WebEnv" ).text
    query_key = root.find( "QueryKey" ).text
    
    uids = [x.text for x in root.findall( "*/Id" ) ]
    
    pubmedlist = []

    params = { 'dbfrom' : 'gds',
               'db' : 'pubmed',
               }
    
    params = urllib.urlencode( params )
    # necessary to preserve 1to1 links
    params += "&" + "&".join( ["id=%s" % x for x in uids ] )
    data = urllib2.urlopen( query_retrieve, params )
        
    etree = ET.parse(data)
    root = etree.getroot()
    
    map_uid2pmid = {}
    for linkset in root.findall("LinkSet"):
        uid = linkset.find( "*/Id" ).text
        try:
            pmid = linkset.find("./LinkSetDb/Link/Id" ).text
        except AttributeError:
            pmid = None
        map_uid2pmid[uid] = pmid

    params = { 'db' : 'gds',
               'id' : ",".join( uids ) }

    params = urllib.urlencode( params )
    data = urllib2.urlopen( query_fetch, params ).read()
    
    map_uid2accession = {}
    map_uid2description = {}
    map_pmid2accession = {}

    for block in data.split( "\n\n" ):
        uid = re.search("ID: (\d+)", block).groups()[0]
        accession = re.search("Accession: (\S+)", block).groups()[0]
        map_uid2accession[uid] = accession
        lines = block.split("\n")
        description = lines[0]
        map_uid2description[uid] = description
        map_pmid2accession[map_uid2pmid[uid]] = accession.encode( "ascii", "ignore") 

    url_pmid = "http://www.ncbi.nlm.nih.gov/pubmed/%s" 
    url_geo = "http://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=%s"

    outfile_uid = IOTools.openFile( "uids.tsv", "w" )
    outfile_uid.write( "uid\tdescription\tpmid\taccession\n" )
    for uid in uids:
        outfile_uid.write( "\t".join( map(str,
                                          ( uid,
                                            map_uid2description[uid],
                                            url_pmid % map_uid2pmid.get( uid, ""),
                                            url_geo % map_uid2accession.get( uid, "") ) ) ) + "\n" )
        

    outfile_pmid = IOTools.openFile( "pmid.tsv", "w" )
    outfile_pmid.write( "pmid\tyear\tjournal\ttitle\tabstract\tgeo\n" )

    E.info( "retrieving pubmed records" )
    # output by pubmed id
    for pmid in map_uid2pmid.values():
        if pmid == None: continue
        print pmid
        # retrieve record

        params = { 'db' : 'pubmed',
                   'id' : pmid,
                   'retmode' : 'xml' }
        
        params = urllib.urlencode( params )
        data = urllib2.urlopen( query_fetch, params )

        etree = ET.parse(data)
        root = etree.getroot()
        article = root.find( "PubmedArticle" )
        assert article != None
        journal = article.find( "*//Journal" )
        assert journal != None
        year = journal.find( "./JournalIssue/PubDate/Year" ).text
        journal_title = journal.find("Title").text

        title = article.find( "*//ArticleTitle" ).text.encode("ascii", "ignore")
        try:
            abstract = article.find( "*//AbstractText" ).text.encode( "ascii", "ignore")
        except AttributeError:
            abstract = ""

        outfile_pmid.write( "\t".join( map(str,
                                           ( url_pmid % pmid,
                                             year,
                                             journal_title,
                                             title,
                                             abstract, 
                                             url_geo % map_pmid2accession.get( pmid, "") ) ) ) + "\n" )

                                             


    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )






################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
align_transcripts.py - multiply align transcripts
=================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script will multiply align transcripts. Transcripts
are aligned as amino acid sequences and later back-translated
to codon sequences.

The script is aware of gene structures will correctly align
exons in sequence. 

Usage
-----

Example::

   python align_transcripts.py --help

Type::

   python align_transcripts.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import optparse
import math

USAGE="""python %s [OPTIONS] < transcripts > filtered

Align peptide/DNA sequences.

If there are multiple transcripts for a gene, concatenate the
exons into one pseudo-transcript and submit that to the multiple
alignment in order to avoid incompatible exons to be aligned.
After alignment, the pseudo-transcripts are split into separate
parts again.

If the peptide sequence contains at least one character ``U``, the
sequence is assumed to be selonoprotein. 

""" % sys.argv[0]

import CGAT.Experiment as E
import CGAT.Genomics as Genomics
import CGAT.Mali as Mali
import CGAT.Exons as Exons
import CGAT.WrapperMuscle as WrapperMuscle
import CGAT.Intervals as Intervals
from peptides2cds import getMapPeptide2Cds
import alignlib
import CGAT.GTF as GTF
import CGAT.IOTools as IOTools

def buildGeneMap( identifiers, separator = "|" ):
    """build map of predictions to genes.

    Use an identifier syntax of species|transcript|gene.
    If none is given, all transcripts are assumed to be 
    from their own gene.
    """

    map_id2gene, map_gene2ids = {}, {}

    for id in identifiers:
        f = id.split(separator)
        if len(f) < 3:
            gene = id
        else:
            gene = f[0] + separator + f[2]
        map_id2gene[id] = gene
        if gene not in map_gene2ids: map_gene2ids[gene] = []
        map_gene2ids[gene].append(id)

    return map_id2gene, map_gene2ids

class Segment:
    def __init__(self):
        pass

#---------------------------------------------------------
#---------------------------------------------------------
#---------------------------------------------------------
## Note: maps are in one-based coordinates, but residue
## return values are in zero-based coordinates
## The calling functions have to deal with this
def MapC2PRight( map_p2c, residue ):
    """mapping function."""
    if residue == map_p2c.getColTo():
        return map_p2c.getRowTo()

    while residue < map_p2c.getColTo():
        x = map_p2c.mapColToRow( residue )
        if x >= 0: return x
        residue += 1

    return -1

def MapP2CRight( map_p2c, residue ):
    """mapping function."""

    if residue == map_p2c.getRowTo():
        return map_p2c.getColTo()

    while residue < map_p2c.getRowTo():
        x = map_p2c.mapRowToCol( residue )
        if x >= 0: return x
        residue += 1

    return -1

def MapP2CLeft( map_p2c, residue ):
    """mapping function."""
    while residue >= 0:
        x = map_p2c.mapRowToCol( residue )
        if x >= 0: return x
        residue -= 1

    return -1

#---------------------------------------------------------
#---------------------------------------------------------
#---------------------------------------------------------

def buildFragments( exons, input, pseudogenes, options, coordinate_factor=3,
                    map_peptide2cds = {},
                    cds_sequences = {},
                    ):
    """build fragments out of a sorted list of overlapping exons.

    The exon list has to be sorted by genome_from, length.

    The algorithm works in the following way::
    
       start                                                                   |end
       |
       =======================================================================
          ====================================================================
              =============================================
          ----                                             -------------------
       ---    ---------------------------------------------

    There are three coordinates for each exon involved:

    -> genomic exon coordinates: fields genome_from, genome_to
    -> coordinates in peptide sequence: fields exon_from, exon_to
    -> coordinates in cds sequence: correspond to exon_from, exon_to,
       unless the transcript contains frameshifts. cds coordinates
       are obtained using the maps in map_peptide2cds.

    If map_peptide2cds is given, it is used for mapping exon boundaries.
    If cds_sequences is given, nucleotide alignments are produces as
    well as aa alignments.
    """


    # intervals
    exon_intervals = map(lambda x: (x.mGenomeFrom, x.mGenomeTo), exons)

    # build intersection set of intervals
    small_intervals = Intervals.getIntersections( exon_intervals )

    nomap = lambda x: x

    if options.loglevel >= 5:
        for e in exons:
            options.stdlog.write( "# exon: %s\n" % str(e))

    if coordinate_factor != 1:

        frames = []
        # collect frames for each interval and assign exons to it
        # both outgoing and in-going frames are checked
        # Frame is the number of residues at start of exon that
        # belong to previous codon.
        
        for start, end in small_intervals:
            ff = {}
            for e in exons:
                if min(end, e.mGenomeTo) - max(start, e.mGenomeFrom) <= 0:
                    continue

                ## collect the three sets of coordinates for this exon
                genome_from, genome_to = e.mGenomeFrom, e.mGenomeTo
                peptide_from, peptide_to = e.mPeptideFrom, e.mPeptideTo
                try:
                    cds_from, cds_to = e.mCdsFrom, e.mCdsTo
                except AttributeError:
                    continue
                    raise "no cds boundaries for %s" % (str(e))
                
                # get cds coordinates of this fragment within exon
                cds_start = cds_from + (start - genome_from)
                cds_end   = cds_to - (genome_to - end) 

                # get map to adjust frame-shifted coordinates
                if e.mQueryToken in map_peptide2cds:
                    map_p2c = lambda x: MapP2CRight( map_peptide2cds[e.mQueryToken], x)
                    map_c2p = lambda x: MapC2PRight( map_peptide2cds[e.mQueryToken], x)                    
                else:
                    map_p2c = nomap
                    map_c2p = nomap
                
                # get peptide coordinates of this fragment within exon
                peptide_start = map_c2p(cds_start)
                peptide_end = map_c2p(cds_end) 
                
                # compute frame based on peptide coordinates
                if peptide_start % coordinate_factor != 0:
                    start_frame = coordinate_factor - peptide_start % coordinate_factor
                else:
                    start_frame = 0

                end_frame = peptide_end % coordinate_factor

                # adjust frame: frame of fragment - original frame
                key = (start_frame, end_frame)
                if key not in ff: ff[key] = []
                ff[key].append(e)
                
            frames.append(ff)

        segments = []
        
        ## build up segments according to exons
        ## iteration over intervals
        for x in range(len(small_intervals)):

            start,end = small_intervals[x]
            l = end - start

            ## iteration over frames within interval
            for key, exons_in_intervall_and_frame in frames[x].items():

                start_frame, end_frame = key
                
                if options.loglevel >= 4:
                    options.stdlog.write("# interval: %i-%i in frame %i:%i\n" % (start, end, start_frame,end_frame) )
                
                segment = Segment()
                segment.mGenomeFrom = start
                segment.mGenomeTo = end
                segment.frame = start_frame
                segment.mMembers = []
                segment.mFragments = []
                segment.mCds = []

                ## iteration over exons with the same frame within an interval
                for e in exons_in_intervall_and_frame:

                    # sanity check - why was this added?
                    if min( end, e.mGenomeTo) - max(start, e.mGenomeFrom) <= 0:
                        raise "within sanity check."
                        continue
                        
                    assert( e.mQueryToken not in segment.mMembers )

                    ## collect the three sets of coordinates for this exon
                    genome_from, genome_to = e.mGenomeFrom, e.mGenomeTo
                    peptide_from, peptide_to = e.mPeptideFrom, e.mPeptideTo
                    cds_from, cds_to = e.mCdsFrom, e.mCdsTo

                    # get cds coordinates of this fragment within exon
                    cds_start = cds_from + (start - genome_from)
                    cds_end   = cds_to - (genome_to - end)

                    # get map to adjust frame-shifted coordinates
                    if e.mQueryToken in map_peptide2cds:
                        map_p2c = lambda x: MapP2CRight( map_peptide2cds[e.mQueryToken], x)
                        map_c2p = lambda x: MapC2PRight( map_peptide2cds[e.mQueryToken], x)                   
                    else:
                        map_p2c = nomap
                        map_c2p = nomap

                    # get peptide coordinates of this fragment within exon
                    peptide_start = map_c2p(cds_start)
                    peptide_end = map_c2p(cds_end)

                    # compute frame based on peptide coordinates
                    #
                    # if frame is not 0, include residue from previous exon. 
                    # Thus: split codon residues are always assigned to the next exon.
                    #
                    # IAW: Frame is the number of residues at start of exon that
                    # belong to previous codon.

                    if start_frame:
                        dfrom = 3 - start_frame
                    else:
                        dfrom, dto = 0, 0

                    dto = (peptide_end - peptide_start + dfrom) % 3

                    x = int( (peptide_start - dfrom) / coordinate_factor)
                    y = int( (peptide_end - dto) / coordinate_factor)

                    if options.loglevel >= 5:
                        options.stdlog.write("# %s: start_frame=%i end_frame=%i dfrom=%i dto=%i x=%i y=%i peptide:%4i-%4i cds:%4i-%4i len=%i len=%i seq=%s\n" % \
                                                 (e.mQueryToken, start_frame, end_frame,
                                                  dfrom, dto, x, y,
                                                  peptide_start, peptide_end,
                                                  cds_start, cds_end,                                                  
                                                  len(input[e.mQueryToken]),
                                                  y-x, 
                                                  input[e.mQueryToken][x:y]))

                    ## assertion only work for genes without frameshifts
                    assert( (peptide_start - dfrom) % coordinate_factor == 0)
                    if (peptide_start - dfrom) % coordinate_factor != 0:
                        pseudogenes.add( e.mQueryToken )

                    assert( (peptide_end - dto) % coordinate_factor == 0)

                    if (peptide_end - dto) % coordinate_factor != 0 :
                        pseudogenes.add( e.mQueryToken )

                    # assert that boundaries work, but ignore errors in last exon or intervals that are smaller than a codon
                    if x < len(input[e.mQueryToken]) and end - start > 3:
                        assert x <= y, "sanity check x <= y failed: x=%i, y=%i, l=%i" % (x, y, len(input[e.mQueryToken]) )

                    # only add meaningfull fragments with non-zero length.
                    if x < y and x < len(input[e.mQueryToken]) and x >= 0:
                        segment.mMembers.append( e.mQueryToken )
                        pep_fragment = input[e.mQueryToken][x:y] 
                        segment.mFragments.append( pep_fragment )

                        ## add nucleotides for the peptide sequence whilst
                        ## skipping over codons containing frameshifts.
                        ## adding "NNN" for skipped codons (identified by
                        ## gaps that are multiples of 3)
                        ## walk along the peptide sequence
                        if cds_sequences and e.mQueryToken in cds_sequences:

                            my_map = map_peptide2cds[e.mQueryToken]
                            # keep case of sequence
                            cds_sequence = cds_sequences[e.mQueryToken]
                            s = []
                            last_i = x * coordinate_factor
                            last_m = my_map.mapRowToCol(x)
                            
                            for peptide_pos in range(x, y):
                                
                                pos = [ my_map.mapRowToCol( i ) for i in range( peptide_pos * coordinate_factor,
                                                                                (peptide_pos + 1) * coordinate_factor) ]
                                
                                ## do not add codons that are incomplete in the peptide sequence
                                if -1 in pos: 
                                    s.append( "NNN" )
                                    continue
                                
                                s += [cds_sequence[x] for x in pos ]

#                             for i in range( y * coordinate_factor, my_map.getRowTo()):
#                                 m = my_map.mapRowToCol(i)
#                                 if m >= 0:
#                                     break

#                             if i - last_i >= 3:
#                                 s.append( "NNN" * int(math.floor( (i-last_i) / 3.0)))
                                
#                             # fill up so that cds sequence is in-frame with nucleotide sequence
                            cds_fragment = "".join(s)

#                             if len(s) % 3 != 0:
#                                  s += "N" * (3 - (len(s) % 3) )

                            segment.mCds.append( cds_fragment )                                

                assert( len(segment.mCds) == len(segment.mFragments) )
                
                ## sanity check: are all sequence segments the same?
                # do not compare boundary residues, as they might be different
                # due to alternative splicing.
                xset = set()
                for x in segment.mFragments:
                    xset.add(x[1:-1])

                ## sequence segments might not be identical
                ## due to different locations of frameshifts within the
                ## sequences. In this case resolve by aligning
                ## the segments and providing the consensus alignment as sequence
                if len(xset) > 1:

                    muscle = WrapperMuscle.Muscle()
                    mali = Mali.Mali()
                    for x in range(len(segment.mFragments)):
                        mali.addSequence( segment.mMembers[x], 0, 0, segment.mFragments[x] )

                    aligned = muscle.Run( mali )

                    pep_consensus = aligned.getConsensus()
                    if options.loglevel >= 6:
                        options.stdlog.write("# consensus peptide alignment:\n")
                        aligned.writeToFile( options.stdlog )
                        options.stdlog.write( pep_consensus + "\n" )
                        
                    ## substitute for consensus
                    segment.mMali = aligned
                    segment.mFragments = [pep_consensus]

                    ## do the same for the nucleotide fragments
                    ## thread each sequence through the peptide alignment
                    if segment.mCds:
                        
##                         muscle = WrapperMuscle.Muscle()
##                         mali = Mali.Mali()
##                         for x in range(len(segment.mCds)):
##                             mali.addSequence( segment.mMembers[x], 0, 0, segment.mCds[x] )
                            
##                         aligned = muscle.Run( mali )

##                         cds_consensus = aligned.getConsensus( mark_with_gaps = True )
                        
##                         if options.loglevel >= 6:
##                             options.stdlog.write("# consensus cds alignment:\n")
##                             aligned.writeToFile( options.stdlog )
##                             options.stdlog.write( cds_consensus + "\n" )
##                             options.stdlog.flush()

                        cds_consensus = [ "-" ] * len(pep_consensus) * 3

                        ## mask inconsistent positions in the consensus peptide
                        ## these are due to frameshifts within an exon
                        columns = segment.mMali.getColumns()
                        for c in range(len(columns)):
                            s = columns[c]
                            counts = [(a, s.count(a)) for a in set(list(s)).difference( set("-") ) ]
                            if len(counts) > 1:
                                cds_consensus[c*3:c*3+3] = ["N"] * 3

                        for c in range(len(segment.mCds)):
                        
                            if options.loglevel >=2:
                                options.stdlog.write("# building map between consensus peptide and %s.\n" % segment.mMembers[c] )
                                options.stdlog.flush()
                            
                            # build map of consensus cds to peptide cds
                            cons_map_p2c = alignlib.makeAlignmentVector()
                            this_cds = segment.mCds[c]
                            try:
                                cons_map_p2c = getMapPeptide2Cds( pep_consensus,
                                                                  this_cds,
                                                                  options )

                            except ValueError, msg:
                                if options.loglevel >= 2:
                                    options.stdlog.write("# Warning: sequence %s not mappable: %s\n" % (segment.mMembers[c],msg) )
                                continue
                            
                            for x in range( cons_map_p2c.getRowFrom(), cons_map_p2c.getRowTo()):
                                y = cons_map_p2c.mapRowToCol( x )
                                if y < 0: continue
                                if cds_consensus[x] == "-":
                                    cds_consensus[x] = this_cds[y]

                        cds_consensus = "".join(cds_consensus)
                        if options.loglevel >= 6:
                            options.stdlog.write("# consensus cds alignment %i, %i:\n" % (len(cds_consensus), len(pep_consensus) ))
                            options.stdlog.write( cds_consensus + "\n" )
                            options.stdlog.flush()
                                
                        ## substitute for consensus
                        segment.mCds = [cds_consensus]

                else:
                    segment.mMali = None
                    
                if len(segment.mFragments) == 0:
                    continue
                
                segment.mSequence = segment.mFragments[0]
                segment.mCdsSequence = segment.mCds[0]

                # assert(len(segment.mSequence) == len(segment.mCdsSequence) / 3 )
                
                segments.append( segment)
                
    return segments

##------------------------------------------------------------
def writeToFile( mali, section, options, is_aligned = True ):
    """write mali to file."""

    outfile = open(options.output_filename_pattern % section, "w" )

    mali.writeToFile(outfile, format=options.output_format )
        
    outfile.close()
        
    if is_aligned and not mali.checkLength():
        raise "mali in file %s has entries of different lengths" % (options.output_filename_pattern % section)
        
##------------------------------------------------------------
if __name__ == '__main__':

    parser = E.OptionParser( version = "%prog version: $Id: align_transcripts.py 2781 2009-09-10 11:33:14Z andreas $", usage = globals()["__doc__"])

    parser.add_option("-m", "--master", dest="master", type="string",
                      help="master sequence."  )

    parser.add_option("-p", "--master-pattern", dest="master_pattern", type="string",
                      help="master pattern."  )

    parser.add_option( "--master-species", dest="master_species", type="string",
                      help="species to use as master sequences."  )

    parser.add_option("-t", "--translate", dest="filename_translation", type="string",
                      help="filename on where to store translated sequences."  )

    parser.add_option("-e", "--exons", dest="filename_exons", type="string",
                      help="filename on where to exon information."  )

    parser.add_option("-g", "--gtf", dest="filename_gtf", type="string",
                      help="filename with exon information in gtf format."  )

    parser.add_option("-c", "--mark-codons", dest="mark_codons", action="store_true",
                      help="mark codons."  )

    parser.add_option( "--remove-stops", dest="remove_stops", action="store_true",
                      help="remove stop codons."  )

    parser.add_option("--mask-stops", dest="mask_stops", action="store_true",
                      help="mask stop codons."  )

    parser.add_option("--mask-char", dest="mask_char", type="string",
                      help="masking character to use."  )

    parser.add_option("-f", "--remove-frameshifts", dest="remove_frameshifts", action="store_true",
                      help="remove columns corresponding to frameshifts."  )

    parser.add_option("-s", "--split-exons", dest="split_exons", action="store_true",
                      help="split columns aligned to different exons in the same gene."  )

    parser.add_option("-a", "--target", dest="target", type="choice",
                      choices=("paml", ),
                      help="perform cleaning up for certain targets." )
    
    parser.add_option("--stop-at", dest="stop_at", type="choice",
                      choices=("aligned", "unaligned", "unpacked"),
                      help="stop at intermediate stage and dump out that alignment.")

    parser.add_option("--force-map", dest="force_map", action="store_true",
                      help="force mapping of sequences that have changed to previous sequence." )

    parser.add_option( "--cds", dest="filename_cds", type="string",
                      help="""filename with cds - useful if you expect pseudogenes in your set. The peptide
sequences will be aligned to the cds sequences. This produces better coordinates."""  )

    parser.add_option("--output", dest="output", type="choice", action="append",
                      choices=("final_aa", "final_na", "aligned_aa", "aligned_na", "all", "unaligned_aa", "unaligned_na", "coords" ),
                      help="which alignment to output: aligned=aligned sequences, but before untangling exons; "
                           " final=final multiple alignment; unaligned=unaligned sequences; "
                           " coords=genomic coordinates (corresponding to 'final_na')." )
    
    parser.add_option("--output-filename-pattern", dest="output_filename_pattern", type="string",
                      help="filename pattern for multiple alignment output files. "
                           " If no --output option is given, stdout is used." )

    parser.add_option("--output-filename-coords", dest="output_filename_coords", type="string",
                      help="filename to output coordinates to." )

    parser.add_option("--output-format", dest="output_format", type="choice",
                      choices=("fasta", "stockholm", "clustal", "plain-fasta") ,
                      help="output format of multiple alignments." )

    parser.add_option("--strict", dest="strict", action="store_true",
                      help="apply strict checking." )
    
    parser.set_defaults(
        gap_char = "-",
        mask_char = "x",
        gap_chars = "-.",
        separator = "|",
        master = None,
        master_species = None,
        filename_translation = None,
        filename_exons = None,
        filename_gtf = None,
        master_pattern = None,
        remove_stops = False,
        mark_codons = False,
        mask_unaligned = False,
        split_exons = False,
        remove_frameshifts = False,
        min_segment_length = 5,
        mask_stops = False,
        target = None,
        sequence_separator = "",
        stop_at = None,
        filename_cds = None,
        force_map = False,
        output_filename_pattern="%s.fasta",
        output_filename_coords="coords.tsv.gz",
        output = [],
        output_format = "fasta",
        max_percent_gaps = 0.1,
        max_gaps = 40,
        strict= False,
        )

    (options, args) = E.Start( parser )

    ########################################################
    if "all" in options.output:
        options.output = ["final_aa", "final_na", "aligned_aa", "aligned_na", "unaligned_aa", "unaligned_na", "coords"]

    ########################################################
    ########################################################
    ########################################################    
    # read unaligned sequences
    input = Mali.Mali()
    input.readFromFile( sys.stdin, format="fasta" )

    all_identifiers = input.getIdentifiers()

    alphabet = input.getAlphabet()

    E.info( "sequences are of alphabet: %s" % alphabet )
        
    if alphabet == "aa":
        coordinate_factor = 3.0

        # build list of selenoproteins
        selenoproteins = set( [x.mId for x in input.values() if "U" in x.mString.upper() ] )

    else:
        coordinate_factor = 1.0
        selenoproteins = ()

    ########################################################
    ########################################################
    ########################################################    
    # read cds sequences
    if options.filename_cds:
        cds_sequences = Genomics.ReadPeptideSequences( open(options.filename_cds, "r") )
        map_peptide2cds = {}

        nskipped = 0
        for x in all_identifiers:

            E.debug( "building map_peptide2cds for sequence %s" % (x) )
                
            if x not in cds_sequences:
                nskipped += 1
                continue
            
            try:
                map_p2c = getMapPeptide2Cds( input[x],
                                             cds_sequences[x],
                                             options )

            except ValueError, msg:
                E.warn( "sequence %s not mappable: %s" % (x,msg) )
                nskipped += 1
                continue

            num_peptide_gaps = len( re.sub("[^-]", "", input[x] ) )
            ngaps = map_p2c.getNumGaps() - (num_peptide_gaps *  3) - abs(len(input[x]*3)-len(cds_sequences[x]))
            
            l = map_p2c.getLength()
            if float(ngaps) / float(l) > options.max_percent_gaps or ngaps > options.max_gaps:
                E.warn( "map between cds to peptide has too many gaps: %s: %i gaps out of length %i" % (x, ngaps, l) )
                nskipped += 1
                continue
            
            map_peptide2cds[x] = map_p2c

            if options.loglevel >= 3:
                f = alignlib.AlignmentFormatEmissions( map_p2c )
                options.stdlog.write("# p2c: " + "\t".join( map(str, (x, str(f),
                                                                      len(input[x]), len(cds_sequences[x])) ) )+"\n")
            

        E.info("built %i maps between peptides and cds - %i skipped" % \
                   (len(map_peptide2cds),
                    nskipped ))
    else:
        map_peptide2cds = {}

    map_id2gene, map_gene2ids = buildGeneMap( all_identifiers )

    ########################################################
    ########################################################
    ########################################################
    ## read exon informmation
    if options.filename_exons or options.filename_gtf:

        # read exon boundaries and keep forward coordinates
        if options.filename_exons:
            exons = Exons.ReadExonBoundaries( open(options.filename_exons, "r"),
                                              filter=set(all_identifiers),
                                              from_zero = True)
        elif options.filename_gtf:
            exons = Exons.ReadExonBoundaries( open(options.filename_gtf, "r"),
                                              filter=set(all_identifiers),
                                              format = "gtf",
                                              gtf_extract_id = re.compile('transcript_id \"(\S+)\";'),
                                              from_zero = True)

            
        ########################################################
        ########################################################        
        ########################################################
        ## Exon boundaries contain the fields mPeptideFrom and mPeptideTo.
        ## These fields correspond to the cds coordinates, not peptide coordinates.
        ## This is no problem, if there are no frameshifts. However, in the presence
        ## of frameshifts, these two coordinates will be different. Thus, if cds
        ## sequences are given, the exon boundaries are corrected, such that
        ## mPeptideFrom and mPeptideTo contain the peptide coordinates and
        ## mCdsFrom and mCdsTo contain the cds coordinates.
        ########################################################
        ########################################################
        ########################################################
        
        ## adjust exon boundaries
        if map_peptide2cds:
            
            # if map_peptide2cds is given, pseudogenes can be treated properly
            E.info("checking exon boundaries." )
                
            nmissing, ndifferences, nstop_codons, ndeleted_empty, nunmappable = 0, 0, 0, 0, 0
            
            # minimum genomic coordinates and strands for a gene
            genome_starts = {}
            # maps of cds sequence to genomic coordinates. These are
            # zeroed and increasing for both forward and reverse strand
            map_cds2genome = {}
            
            for key, ee in exons.items():

                if key not in map_peptide2cds:
                    nmissing += 1
                    continue

                map_p2c = map_peptide2cds[key]

                ## a patch to eliminate empty last exons
                if ee[-1].mPeptideTo == ee[-1].mPeptideFrom and \
                   ee[-1].mGenomeTo == ee[-1].mGenomeFrom:
                    E.warn("%s of length %i: deleting empty last exon: %s." % \
                               (key, len(input[key]), str(ee[-1])))
                    del ee[-1]
                    ndeleted_empty += 1
                    
                if ee[-1].mPeptideTo != map_p2c.getColTo():

                    E.debug( "%s" % str(ee[-1]) )
                    E.warn( "%s of length %i: peptide and exon do not correspond: %i <> %i" %\
                                (key, len(input[key]), ee[-1].mPeptideTo, map_peptide2cds[key].getColTo()) )
                    ndifferences += 1
                    d = ee[-1].mPeptideTo - map_peptide2cds[key].getColTo()
                    if d == 3:
                        E.warn("%s: assuming difference is stop codon - exon shortened." )
                        nstop_codons += 1
                    elif d > 0:
                        E.warn("%s: fixing difference of %i nucleotides - incomplete stop-codon?" % (key,d) )
                    else:
                        # if the exon information extends to higher residues than the map_peptide2cds
                        raise ValueError("not able to fix exon boundary incongruence: %i != %i." %\
                                             (ee[-1].mPeptideTo, map_p2c.getColTo()))
                    
                    old_peptide_end, old_genome_end = ee[-1].mPeptideTo, ee[-1].mGenomeTo 
                    ee[-1].mPeptideTo -= d
                    ee[-1].mGenomeTo -= d

                    d = ee[-1].mPeptideTo - ee[-1].mPeptideFrom
                    if d <= 0:
                        del ee[-1]
                        ee[-1].mPeptideTo += d
                        ee[-1].mGenomeTo += d

                    E.debug( "%s: fixed exon end from %i to %i (%i to %i)" % \
                                 (key, old_peptide_end, ee[-1].mPeptideTo,
                                  old_genome_end, ee[-1].mGenomeTo ) )

                is_negative_strand = ee[0].mSbjctStrand == "-"
                # note that exon coordinates are already inverted
                # and negative strand coordinates are negative
                # Thus the following works for both forward and reverse strand
                genome_start = ee[0].mGenomeFrom
                genome_starts[key] = (is_negative_strand, genome_start)

                map_c2g = alignlib.makeAlignmentBlocks()

                for e in ee:
                    # map boundaries
                    # note: map_p2c is in 1 based coordinates
                    peptide_from = MapC2PRight(map_p2c, e.mPeptideFrom )
                    peptide_to = MapC2PRight(map_p2c, e.mPeptideTo)

                    if peptide_from < 0 or peptide_to < 0:
                        E.debug( "%s" % str(e) )
                        E.warn( "%s of length %i: exon boundary could not be mapped: from %i->%i to %i->%i" %\
                                    (key, len(input[key]),
                                     e.mPeptideFrom, peptide_from,
                                     e.mPeptideTo, peptide_to))
                        E.debug("%s" %str(alignlib.AlignmentFormatEmissions( map_p2c )) )
                        nunmappable += 1
                        continue

                    e.mCdsFrom = e.mPeptideFrom
                    e.mCdsTo = e.mPeptideTo
                    e.mPeptideFrom = peptide_from
                    e.mPeptideTo = peptide_to

                    # build map of cds to genomic sequence
                    map_c2g.addDiagonal( e.mCdsFrom, e.mCdsTo, e.mGenomeFrom - genome_start - e.mCdsFrom ) 
                    map_cds2genome[key] = map_c2g

            E.info("checked exon boundaries against cds: missing=%i, differences=%i, fixed_stops=%i, deleted_empty=%i, nunmappable=%i" %\
                                     (nmissing, ndifferences, nstop_codons, ndeleted_empty, nunmappable))

        else:
            E.info("no checking of exon boundaries - assumed to be correct.")
            E.info("removing stop codons.")
            #
            # remove stop codons from exons
            #
            # Tests whether the length of the peptide sequence and
            # the exon correspond. If there is a difference, truncate.
            # Note: this fails for pseudogenes. Thus use peptide2cds
            # to map these
            for key, ee in exons.items():

                reference_length = 3 * len(input[key])

                if ee[-1].mPeptideTo > reference_length:
                    ee[-1].mPeptideTo -= 3
                    ee[-1].mGenomeTo -= 3

                if ee[-1].mPeptideTo - ee[-1].mPeptideFrom == 0:
                    del ee[-1]

                for e in ee:
                    e.mCdsFrom = e.mPeptideFrom
                    e.mCdsTo = e.mPeptideTo

            map_peptide2cds = {}

        E.debug( "read exons for %i sequences." % len(exons) )

    else:
        # if no exons are given, all transcripts are assumed to be single exon
        # genes
        exons = {}
        for gene, identifiers in map_gene2ids.items():
            for id in identifiers:
                e = Exons.Exon()
                e.mQueryToken = id
                e.mPeptideFrom = map_peptide2cds[id].getRowFrom()
                e.mPeptideTo = map_peptide2cds[id].getRowTo()
                e.mGenomeFrom = map_peptide2cds[id].getColFrom()
                e.mGenomeTo = map_peptide2cds[id].getColTo()
                e.mCdsFrom = e.mGenomeFrom
                e.mCdsTo = e.mGenomeTo
                exons[id] = [e]
                
    ##########################################################################
    ##########################################################################
    ##########################################################################
    ## Build the packed alignment
    ##########################################################################        
    unaligned = Mali.Mali()
    unaligned_cds = Mali.Mali()
    
    map_gene2fragments = {}

    pseudogenes = set()
    nwarnings_length, nwarnings_sequence = 0, 0
    
    nwarnings_empty = 0
    
    for gene, identifiers in map_gene2ids.items():
        
        ## collect all exons and sort them by genomic location
        exon_list = []
        is_seleno = False
        for id in identifiers:
            exon_list += exons[id]
            is_seleno |= id in selenoproteins

        # sort such that the larger exons come first
        # if they start at the same residue
        exon_list.sort( lambda x,y: cmp( (x.mGenomeFrom, -x.mGenomeTo), (y.mGenomeFrom, -y.mGenomeTo)) )

        e = exon_list[0]

        max_to = e.mGenomeTo
        overlapping_exons = [ e ]
        fragments = []
        for e in exon_list[1:]:

            # ignore empty exons
            # Example for empty exon: ENSP00000344726, where the last exon
            # just consists of the stop codon. Should be filtered out, but
            # you never know.
            if e.mPeptideFrom == e.mPeptideTo: continue

            if max_to <= e.mGenomeFrom:
                # no overlap, process chunks
                fragments += buildFragments( overlapping_exons, input, 
                                             pseudogenes, options,
                                             map_peptide2cds = map_peptide2cds,
                                             cds_sequences = cds_sequences )
                overlapping_exons = []

            overlapping_exons.append( e )
            max_to = max( e.mGenomeTo, max_to )

        fragments += buildFragments( overlapping_exons, input, pseudogenes, options,
                                     map_peptide2cds = map_peptide2cds,
                                     cds_sequences = cds_sequences )
        
        map_gene2fragments[gene] = fragments

        # build unaligned sequence
        sequence = options.sequence_separator.join( map( lambda x: x.mSequence, fragments) )

        # sanity check if cds sequence is given: does it correspond to peptide sequence?
        if cds_sequences:
            cds_sequence = options.sequence_separator.join( map( lambda x: x.mCdsSequence, fragments) )
            mapped_sequence = Genomics.translate( cds_sequence, is_seleno = is_seleno )
            if mapped_sequence != sequence:
                if len(mapped_sequence) == len(sequence):
                    ncounts = 0
                    for x in range(len(mapped_sequence)):
                        if mapped_sequence[x].upper() != sequence[x].upper():
                            E.info("gene %s: amino acid position %i has changed from %s to %s" % (gene, x, sequence[x], mapped_sequence[x]))
                            ncounts += 1
                    if ncounts > 0:
                        E.warn( "gene %s: %i amino acid positions have changed" % (gene, ncounts))
                        nwarnings_sequence += 1
                else:
                    E.warn("gene %s: back-translated sequence is different from original" % (gene))
                    E.warn("      Original : %s" % (sequence))
                    E.warn("      Backtrans: %s" % (mapped_sequence))
                    E.warn("assumed to be pseudogenes and ignored." )
                    nwarnings_sequence += 1

                    if options.strict:
                        raise ValueError("gene %s: back-translated sequence has different length from original\n" % (gene))

        unaligned.addSequence( gene, 0, len(sequence), sequence )
        unaligned_cds.addSequence( gene, 0, len(cds_sequence), cds_sequence )
        
    if "unaligned_aa" in options.output:
        writeToFile( unaligned, "unaligned_aa", options, is_aligned = False )

    if "unaligned_na" in options.output:
        writeToFile( unaligned_cds, "unaligned_na", options, is_aligned = False )

    if options.stop_at == "unaligned":
        unaligned.writeToFile( options.stdout, format=options.output_format )
        E.Stop()
        sys.exit(0)

    ##########################################################################
    ##########################################################################
    ##########################################################################
    ## Perform the multiple alignment
    ##########################################################################        
    muscle = WrapperMuscle.Muscle()

    aligned = muscle.Run( unaligned )

    # substitute U for X in selenoproteins
    # note that this is a greedy substitution - genuine stop-codons
    # will be overridden as well.
    if selenoproteins:
        for gene_id, s in aligned.items():
            for pid in map_gene2ids[gene_id]:
                if pid in selenoproteins:
                    s.mString = s.mString.replace( "X", "U" ).replace( "x", "u")
                    break

    if "aligned_aa" in options.output:
        writeToFile( aligned, "aligned_aa", options)

    # perform sanity checks
    # all aligned sequences should have the same length
    # all sequences should be identical to the unaligned sequences
    width = aligned.getWidth()
    for key,val in aligned.items():
        if len(val.mString) != width:
            raise ValueError("incompatible lenghts in %s: %i should be %i." % (key, len(val.mString), width ))
        try:
            s1 = re.sub( "[%s]" % options.gap_chars, "", unaligned[key] )
        except KeyError:
            continue
        s2 = re.sub( "[%s]" % options.gap_chars, "", val.mString )
        if s1.upper() != s2.upper():
            raise ValueError("sequence changed in %s:\nold=%s\nnew=%s" % (key, s1, s2) )

    if options.stop_at == "aligned":
        aligned.writeToFile( options.stdout, format=options.output_format)
        E.Stop()
        sys.exit(0)

    ##########################################################################
    ##########################################################################
    ##########################################################################
    ## output the packed alignment as nucleotides 
    ## The output does not contain any frameshifts that might have been
    ## present in the original sequences
    if "aligned_na" in options.output:

        aligned_cds = Mali.Mali()

        for id in aligned.getIdentifiers():
            entry = aligned.getEntry(id)

            s = []
            c = 0
            gc = aligned.mGapChar
            cds_sequence = unaligned_cds[id]
            for x in range(len(entry.mString)):
                if entry.mString[x] == gc:
                    s.append( gc * 3 )
                else:
                    s.append( cds_sequence[c:c+3] )
                    c += 3

            aligned_cds.addSequence( id, entry.mFrom * 3, entry.mTo * 3, "".join(s) )

        writeToFile( aligned_cds, "aligned_na", options )
        
    ##########################################################################
    ##########################################################################
    ##########################################################################
    ## unpack the protein level alignment
    ## columns stay the same, but alignments with multiple transcripts per gene
    ## are unpacked.
    ##########################################################################        
    unpacked = Mali.Mali()

    for id in aligned.getIdentifiers():
        
        gene = id
        sequence = aligned[id]
        entry = aligned.getEntry(id)

        fragments = map_gene2fragments[gene]
        ## split aligned sequence in its constituent parts
        transitions = []
        c = 0
        for x in fragments:
            c += len(x.mSequence)
            transitions.append(c)

        segments = entry.getSegments( transitions )

        if len(segments) != len(transitions):

            for x in range(min(len(segments), len(transitions))):

                segment = sequence[segments[x][0]:segments[x][1]]

                if options.loglevel >= 8:
                    options.stdlog.write("# %s: transition=%i segment=%i-%i members=%s\n# %s: sequence=%s segment=%s\n" % (id,
                                                                                                                           transitions[x],
                                                                                                                           segments[x][0],
                                                                                                                           segments[x][1],
                                                                                                                           fragments[x].mMembers,
                                                                                                                           id,
                                                                                                                           fragments[x].mSequence,
                                                                                                                           segment ) )



            print "# %s: nfragments=%i" % (id, len(fragments))
            print "# %s: nsegments=%i segments=%s lsequence=%i" % (id, len(segments), str(segments), len(sequence))
            print "# %s: ntransitions=%i transitions=%s lgene=%i" % (id, len(transitions), transitions, len(unaligned[gene]))
            print sequence
            print unaligned[gene]

        assert (len(segments) == len(transitions))

        sequences = {}          
        for member in map_gene2ids[gene]:
            sequences[member] = []

        fragments = map_gene2fragments[id]

        for x in range(len(segments)):

            segment = sequence[segments[x][0]:segments[x][1]]

            if options.loglevel >= 8:
                options.stdlog.write("# %s: transition=%i segment=%i-%i members=%s\n# %s: sequence=%s segment=%s\n" % (id,
                                                                                                                       transitions[x],
                                                                                                                       segments[x][0],
                                                                                                                       segments[x][1],
                                                                                                                       fragments[x].mMembers,
                                                                                                                       id,
                                                                                                                       fragments[x].mSequence,
                                                                                                                       segment ) )


            added = set()

            ## if it was a consensus string, deconvolute with multiple alignment
            if fragments[x].mMali:
                for member in fragments[x].mMembers:
                    s1 = fragments[x].mMali.getSequence( member ).mString
                    s = []
                    i = 0
                    for ch in segment:
                        if ch in options.gap_chars:
                            s.append( ch )
                        else:
                            s.append(s1[i])
                            i += 1
                    sequences[member].append( "".join(s) )
                    added.add( member )
            else:
            ## simply add the segment
                for member in fragments[x].mMembers:
                    sequences[member].append( segment )
                    added.add( member )

            segment = options.gap_char * (len(segment))

            for member in map_gene2ids[gene]:
                if member not in added:
                    sequences[member].append( segment )

        for member in map_gene2ids[gene]:
            s = "".join(sequences[member])

            if len(re.sub("[%s]" % options.gap_chars, "", s )) == 0:
                E.warn("empty sequence for %s." % member )
                nwarnings_empty += 1
                continue    

            unpacked.addSequence( member, 0, -1, s )

    if options.stop_at == "unpacked":
        aligned.writeToFile( options.stdout, format=options.output_format)
        E.Stop()
        sys.exit(0)

    # perform sanity checks
    # all aligned sequences should have the same length
    # all sequences should be identical to the unaligned sequences
    #   except: sometimes at split codons exon boundaries might
    #   be chosen differently. Thus: allow single residue changes
    
    width = unpacked.getWidth()
    nsubstitutions = 0
    
    for key,val in unpacked.items():
        
        if len(val.mString) != width:
            raise ValueError("incompatible mali lengths in %s: %i should be %i." % (key, len(val.mString), width ))
        
        try:
            sold = re.sub( "[%s]" % options.gap_chars, "", input[key] )
        except KeyError:
            continue
        
        snew = re.sub( "[%s]" % options.gap_chars, "", val.mString )

        if len(sold) != len(snew):
            if key in pseudogenes:
                E.warn("pseudogenic sequence %s changed in length" % (key))
                E.warn("%s:\nold=%s\nnew=%s" % (key, sold, snew))
                nwarnings_length += 1
            elif (len(snew) - len(sold) == 1 and sold == snew[:-1]) or \
                 (len(sold) - len(snew) == 1 and snew == sold[:-1]):
                E.warn("sequence %s changed in length - last residue lost" % (key))
                E.warn("%s:\nold=%s\nnew=%s" % (key, sold, snew))
                nwarnings_length += 1
            elif not options.force_map:
                raise ValueError("sequence changed in length %s:\nold=(%i)\n%s\nnew=(%i)\n%s" % (key, len(sold), sold, len(snew), snew))

        else:
            ## for those with equal length: fix single residue substitutions
            nchanged = 0

            if sold != snew:
                ## fix single character changes
                c = 0
                s = list(val.mString)
                for x in range(len(s)):
                    if s[x] in options.gap_chars:
                        continue
                    if s[x].upper() != sold[c].upper():
                        if options.loglevel >= 8:
                            options.stdlog.write("# %s: fixing split codon: %i->%i %s:%s:%s -> %s:%s:%s\n" %\
                                                     (key, x, c, 
                                                      sold[c-5:c], sold[c], sold[c+1:c+6],
                                                      "".join(s[x-5:x]), s[x], "".join(s[x+1:x+6:])) )
                        nchanged += 1
                        s[x] = sold[c]

                    c += 1

                val.mString = "".join(s)

            if nchanged > 0:
                E.info( "%s: fixed %i split codons." % (key, nchanged) )

            # if more substitutions made than exon boundaries
            # it is likely a programming error
            # should be >=, but was too strict for one fly sequence, so I relaxed it.
            if nchanged > len( exons[key] ):
                raise ValueError("more codons fixed than intron-exon boundaries in sequence %s: %i>=%i" % (key, nchanged, len(exons[key])))
            
        snew = re.sub( "[%s]" % options.gap_chars, "", val.mString )
        
        if sold.upper() != snew.upper():
            if key in pseudogenes:
                E.warn("pseudogenic sequence %s changed." % (key))
                E.warn("%s:\n# old=%s\n# new=%s" % (key, sold, snew))
                nwarnings_sequence += 1
                
            if options.force_map:
                map_old2new = alignlib.makeAlignmentVector()
                alignator = alignlib.makeAlignatorDPFull( alignlib.ALIGNMENT_GLOBAL, -10.0, -1.0 )
                s1 = alignlib.makeSequence( sold )
                s2 = alignlib.makeSequence( snew )
                alignator.align( map_old2new, s1, s2 )

                val.threadSequence( snew, map_old2new )
                E.warn("sequence threaded onto original sequence: %s" % (key))
                nsubstitutions += 1

            else:    
                raise ValueError("# sequence changed in %s:\n# old=%s\n# new=%s" % (key,
                                                                                    sold,
                                                                                    snew))

    if not options.output: 
        unpacked.writeToFile( options.stdout, format=options.output_format )
    else:
        if "final_aa" in options.output:
            writeToFile( unpacked, "final_aa", options )

        if "final_na" in options.output:

            ## map alignment to original cds
            ## frameshifts in the original cdna are ignored
            ## output table with coordinates
            unpacked_cds = Mali.Mali()
            
            info, all_coords = [], []
            
            for key in unpacked.getIdentifiers():

                entry = unpacked.getEntry(key)
                
                p = unpacked[key]
                c = cds_sequences[key]
                
                ###############################
                # build map of cds to alignment position (map_cds2pos)
                try:
                    E.debug("building map for %s" % key)
                    map_pos2cds = getMapPeptide2Cds( p, c,
                                                     options )

                except ValueError, msg:
                    E.warn("final alignment not mappable for %s -skipped" % key )
                    nskipped += 1
                    continue


                #################################
                # map alignment pos to genome
                map_c2g = map_cds2genome[key]
                map_pos2genome = alignlib.makeAlignmentBlocks()
                alignlib.combineAlignment( map_pos2genome, map_pos2cds, map_c2g, alignlib.CR )

                coords = []            
                is_negative_strand, start = genome_starts[key]
                for x in range( len(p) * 3 ):
                    y = map_pos2genome.mapRowToCol(x)
                    if y < 0:
                        coords.append( "" )
                    else:
                        if is_negative_strand:
                            coords.append( 0 - (start + y) )
                        else:
                            coords.append( start + y )
                all_coords.append( coords )
                info.append( ( key, exons[key][0].mSbjctToken, exons[key][0].mSbjctStrand ) )

                ################################
                # create threaded alignment string
                map_pos2cds.switchRowCol()
                map_cds2pos = map_pos2cds

                alignatum = alignlib.makeAlignatum( cds_sequences[key] )

                alignatum.mapOnAlignment( map_cds2pos, len(p) * 3 )
                s = alignatum.getString()
                if len(s) != len(p) * 3:
                    raise ValueError("incomplete aligned string for %s: %s, cds=%s" % (key, s, c ))
                
                unpacked_cds.addSequence( key, 0, len(c), s )

            writeToFile( unpacked_cds, "final_na", options )

            if "coords" in options.output:
                E.info("output genomic coordinates to %s" % options.output_filename_coords)
                with IOTools.openFile( options.output_filename_coords, "w") as outfile:
                    outfile.write( "position\t%s\n" % "\t".join( [ "|".join( map(str,i) ) for i in info ] ) )
                    all_coords = zip( *all_coords )
                    for x, c in enumerate( all_coords ):
                        outfile.write("%i\t%s\n" % (x, "\t".join(map(str, c) ) ) )
                    
    E.info( "ninput=%i, noutput=%i, nskipped=%i, nwarnings_length=%i, nwarnings_sequences=%i, npseudogenes=%i, nsubstitutions=%i" % \
                (input.getLength(),
                 unpacked.getLength(),
                 nskipped,
                 nwarnings_length,
                 nwarnings_sequence,
                 len(pseudogenes),
                 nsubstitutions))
    
    E.Stop()

    




################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
annotator2tsv.py - analyse annotator results
========================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script takes annotator results, parses them, 
filters by FDR and outputs the results 

Usage
-----

Example::

   python annotator2tsv.py --help

Type::

   python annotator2tsv.py --help

for command line help.

Documentation
-------------

Code
----

'''
import sys
import string
import re
import optparse
import collections
import bisect
import os

USAGE="""python %s [OPTIONS] input1 input2

parse annotator files and compute overlap.

Note: this currently does a brute force overlap. In the future
use bme or rtree indices.

Version: $Id: annotator2tsv.py 2861 2010-02-23 17:36:32Z andreas $
""" % sys.argv[0]

import CGAT.Experiment as E
import CGAT.GTF as GTF
import numpy
import CGAT.Stats as Stats

class Identifier:
    def __init__( self, line = None):
        if line: self.read(line)
            
    def read( self, line ):
        data = re.split("\s", line[:-1])
        assert data[0] == "##Id", "formatting error in line %s" % line
        self.mId = int(data[1])
        self.contig = data[2]
        self.mSegments = map( lambda x: tuple(map(int, x[1:-1].split(","))), data[3:] )

    def __str__(self):
        return "##Id\t%s\t%s\t%s" % (self.mId, 
                                     self.contig, 
                                     "\t".join( ["(%i,%i)" % x for x in self.mSegments] ) )

class Workspace:
    def __init__( self, line = None):
        if line: self.read(line)
            
    def read( self, line ):
        data = re.split("\s", line[:-1])
        assert data[0] == "##Work", "formatting error in line %s" % line
        self.contig = data[1]
        self.mSegments = map( lambda x: tuple(map(int, x[1:-1].split(","))), data[2:] )

    def __str__(self):
        return "##Work\t%s\t%s" % (self.contig, "\t".join( ["(%i,%i)" % x for x in self.mSegments] ) )

class Segments:
    def __init__( self, line = None):
        if line: self.read(line)
            
    def read( self, line ):
        data = re.split("\s", line[:-1])
        assert data[0] == "##Segs", "formatting error in line %s" % line
        self.contig = data[1]
        self.mSegments = map( lambda x: tuple(map(int, x[1:-1].split(","))), data[2:] )
    def __str__(self):
        return "##Segs\t%s\t%s" % (self.contig, "\t".join( ["(%i,%i)" % x for x in self.mSegments] ) )

class Annotation:
    def __init__( self, line = None):
        if line: self.read(line)
            
    def read( self, line ):
        data = re.split("\t", line[:-1])
        assert data[0] == "##Ann", "formatting error in line %s" % line
        self.mId = data[1]
        # deal with empty annotations
        if data[2] == "": 
            self.mIdentifiers = []
            return

        try:
            self.mIdentifiers = map(int, data[2:])
        except ValueError, msg:
            raise ValueError( "formatting error in line %s: %s" % (line, msg))

    def __str__(self):
        return "##Ann\t%s\t%s" % (self.mId, "\t".join(map(str, self.mIdentifiers)) )

def readSegments( infile ):
    segments = []
    for line in infile:
        if line.startswith( "##Segs" ):
            segments.append( Segments( line=line ) )


    return segments

def readWorkspace( infile ):
    workspace = []
    for line in infile:
        if line.startswith( "##Work" ):
            workspace.append( Workspace( line=line ) )

    return workspace

def readAnnotations( infile ):
    identifiers = []
    annotations = []
    for line in infile:
        if line.startswith( "##Id" ):
            identifiers.append( Identifier( line=line ) )
        elif line.startswith( "##Ann" ):
            annotations.append( Annotation( line = line ) )    

    map_query2annotation, map_annotation2identifier = {}, {}
    for identifier in identifiers:
        map_annotation2identifier[identifier.mId] = identifier
    for annotation in annotations:
        map_query2annotation[annotation.mId] = annotation

    return identifiers, annotations, map_query2annotation, map_annotation2identifier

def doQuery( options, args):
    """query annotator input for overlapping segments."""

    options.queries.extend( args )

    identifiers, annotations, map_query2annotation, map_annotation2identifier = readAnnotations( open( options.filename_annotations, "r") )
    segments = readSegments( open( options.filename_segments, "r") )
    workspace = readWorkspace( open( options.filename_workspace, "r") )

    if options.filename_segments_gtf:
        with open(options.filename_segments_gtf, "r") as infile:
            segments_gtfs = GTF.readAsIntervals( GTF.iterator( infile ), with_records = True )
    else:
        segments_gtfs = {}

    if options.filename_annotations_gtf:
        with open(options.filename_annotations_gtf, "r") as infile:
            annotations_gtfs = GTF.readAsIntervals( GTF.iterator( infile ), with_records = True )
    else:
        annotations_gtfs = {}

    E.info( "identifiers=%i, annotations=%i, segments=%i, workspace=%i" % (len(identifiers), len(annotations), len(segments), len(workspace)) )

    if options.loglevel >= 3:
        for x in segments: options.stdlog.write( str(x) + "\n" )
        for x in workspace: options.stdlog.write( str(x) + "\n" )
        for x in annotations: options.stdlog.write( str(x) + "\n" )
        for x in identifiers: options.stdlog.write( str(x) + "\n" )

    if options.output_format == "table":
        options.stdout.write("query\tcontig\tanno_start\tanno_end\tanno_id\tseg_start\tseg_end\tgene_id\n" )

    if "all" in options.queries:
        options.queries = map_query2annotation.keys()
        E.debug( "doing all queries: %i" % len(options.queries) )

    for query in options.queries:

        annotation = map_query2annotation[query] 

        # get segments belonging to annotation
        annotation_segments = []
        for id in annotation.mIdentifiers:
            identifiers = map_annotation2identifier[id]
            for start, end in identifiers.mSegments:
                annotation_segments.append( (identifiers.contig, start, end) )

        E.info( "annotation_segments=%s"% len(annotation_segments ) )
        E.debug( "annotation_segments=%s"% str(annotation_segments ) )
        # quick and dirty overlap with workspace
        workspace_overlaps = []
        for contig, start, end in annotation_segments:
            finished = False
            for wrk in workspace:
                if wrk.contig != contig: continue
                for wrk_start, wrk_end in wrk.mSegments:
                    if min(wrk_end, end) - max(wrk_start, start) > 0:
                        workspace_overlaps.append( (contig, start, end ) )
                        finished = True
                        break
                if finished: break

        E.info( "workspace_overlaps=%s"% len(workspace_overlaps ) )
        E.debug( "workspace_overlaps=%s" % str(workspace_overlaps) )

        # Quick and dirty overlap with segments
        overlaps = []
        for contig, start, end in workspace_overlaps:
            for segment in segments:
                if segment.contig != contig: continue
                for seg_start, seg_end in segment.mSegments:
                    if min(seg_end, end) - max(seg_start, start) > 0:
                        overlaps.append( (segment.contig, start, end, seg_start, seg_end) )

        for contig, anno_start, anno_end, seg_start, seg_end in overlaps:
            seg_gene_ids = []
            if contig in segments_gtfs:
                for start,end,gtf in segments_gtfs[contig]:
                    if seg_start == start and seg_end == end:
                        seg_gene_ids.append( gtf.gene_id )

            anno_gene_ids = []
            if contig in annotations_gtfs:
                for start,end, gtf in annotations_gtfs[contig]:
                    if anno_start == start and anno_end == end:
                        anno_gene_ids.append( gtf.gene_id )

            if options.output_format == "table":
                options.stdout.write( "%s\t%s\t%i\t%i\t%s\t%i\t%i\t%s\n" % \
                                          (query, contig, 
                                           anno_start, anno_end, 
                                           ";".join(anno_gene_ids),
                                           seg_start, seg_end, 
                                           ";".join(seg_gene_ids) ) )
            elif options.output_format == "gff":
                options.stdout.write( "\t".join( (contig, 
                                                  "anno", 
                                                  "anno",
                                                  str(seg_start + 1), 
                                                  str(seg_end), 
                                                  ".",
                                                  "+",
                                                  ".",
                                                  'annotation_id = "%s" % query') ) + "\n" )
                
class AnnotatorTest:
    def __init__(self):
        self.mFDR = 0
        self.mQValue = 1

class DataFDR:
    def __init__(self):
        pass

def readAnnotator( infile,
                   delims = "",
                   ignore = "" ):
    """parse annotator output.

    returns (data, fdr).
    """

    kw = "Iteration"
    data = []
    annotator_fdr = {}
    annotator_level = None

    synonyms = {}
    input_files = {}
    
    pvalues, qvalues = [], []

    for line in infile:

        # skip meaningless lines
        if line.startswith("Iteration"): continue       # skip iteration comment lines
        if line.startswith( "Z" ): continue            # skip header line of results
        if len(line) == 1: continue                  # skip trailing blank lines

        if line.startswith( "# Found duplicated annotation:" ):
            term = re.match( "# Found duplicated annotation: (.*)$", line[:-1] ).groups()[0]
            synonyms[term] = []
            continue
        
        if line.startswith( "#  Removing:" ):
            synonym = re.match( "#  Removing:\s+(.*)$", line[:-1] ).groups()[0]
            synonyms[term].append( synonym )
            continue

        if line.startswith( "# Reading file" ):
            filename, option = re.search( "# Reading file (\S+) for option (\S+)", line ).groups()
            filename = os.path.basename( filename )
            if option.startswith("workspace"):
                if "workspace" not in input_files:
                    input_files["workspace"] = filename
                else:
                    input_files["workspace"] += "-%s" % filename
            else:
                input_files[option] = filename

        # parse FDR
        # annotator only reports the FDR for fixed p-value intervals.
        if line.startswith( "--" ):
            if line.startswith("-- False"):
                annotator_level = float(re.search( "-- False Discovery summary for p-value (.+):", line ).groups()[0])
                annotator_fdr[annotator_level] = {}
            elif line.startswith("--  Category"):
                pass
            else:
                if re.search("insufficiently", line): continue
                dd = re.split("\s+", line[4:-1])

                d = DataFDR()
                d.mObserved, d.mAverage, d.mMedian, d.m95 = map(float,dd[1:])
                if d.mObserved > 0:
                    d.mQValue = d.mAverage / d.mObserved
                else:
                    d.mQValue = 1.0
                annotator_fdr[annotator_level][dd[0]] = d
                if dd[0] == "Significant":
                    pvalues.append( annotator_level )
                    qvalues.append( d.mQValue )

        if len(line[:-1].split('\t')) != 9: continue # HACK: accounts for a bug in Annotator output


        try:
            (z, percentchange, pvalue, observed, expected, low95, up95, stddev, description) = line[:-1].split('\t')[:9]
        except ValueError:
            raise ValueError("# parsing error in line: %s" % line[:-1])
        
        # remove leading .
        if description.startswith("."): description = description[1:]

        pvalue = float( pvalue )

        
        d = AnnotatorTest()
        d.mAnnotation = description
        d.mPValue = pvalue
        d.mFoldChange = 1.0 + float(percentchange) / 100.0
        d.mObserved = float(observed)
        d.mExpected = float(expected)
        d.mCI95 = float(low95), float(up95)
        d.mStdDev = float(stddev)
        d.mQValue = 1.0

        ## apply filters
        for c in delims:
            d.mAnnotation = d.mAnnotation.split(c)[0]
        for c in ignore:
            d.mAnnotation = d.mAnnotation.replace(c, '')
            
        data.append(d)

    # sort by increasing value
    pvalues.reverse()
    qvalues.reverse()
    # compute fdr
    # the q-value of each observation is set to the
    # qvalue for the next highest P-Value.
    if len(pvalues) == 0:
        return data, annotator_fdr, synonyms, input_files

    max_pvalue = pvalues[-1]
    
    for d in data:
        pvalue = d.mPValue
        if pvalue > max_pvalue: d.mQvalue = 1.0
        else: d.mQValue = qvalues[max(0, bisect.bisect(pvalues, pvalue) - 1)]

    return data, annotator_fdr, synonyms, input_files

def doOldFDR( options, args ):
    """apply fdr to output of annotator."""

    # read input
    annotators = []
    for filename in args:
        infile = open(filename,"r")
        annotators.append( readAnnotator( infile ))
        infile.close()
        
    # apply filters and create diagnostic plots    
    for filename, data in zip(args, annotators):
        ninput = len(data)
        pvalues = [x.mPValue for x in data]
        vlambda = numpy.arange( 0, max( pvalues ), 0.05 )
        try:
            qvalues = Stats.doFDR( pvalues, vlambda = vlambda, fdr_level = options.fdr )
        except ValueError, msg:
            E.warn( "%s: fdr could not be computed - no filtering: %s" % (filename, msg) )
            continue

        qvalues.plot( filename + "_diagnostics.png" )

        data = [ x[0] for x in zip( data, qvalues.mPassed ) if x[1] ]

def doFDR( options, args ):

    # read input
    annotators = []
    for filename in args:
        infile = open(filename,"r")
        annotators.append( readAnnotator( infile ))
        infile.close()

    do_filter = options.fdr_qvalue != None
    
    extra_headers = set()
    for data, fdr, synonyms, input_files in annotators:
        for key, value in input_files.iteritems():
            extra_headers.add( key )
    extra_headers=sorted(list(extra_headers))

    # note: id used to be file
    options.stdout.write( "id\tover\tcategory\tpvalue\tfold\tobserved\texpected\tci95low\tci95high\tstddev\tfdr\tqvalue\t%s\n" %\
                              "\t".join( extra_headers ))

    # apply filters and create diagnostic plots    
    for filename, vv in zip(args, annotators):
        data, fdr, synonyms, input_files = vv

        ninput = len(data)

        E.info( "processing %s with %i data points" % (filename, ninput) )
        no_fdr = False

        if options.fdr_method in ("annotator", "annotator-estimate"):
            pvalues = fdr.keys()
            pvalues.sort()
            pvalues.reverse()
            for pvalue in pvalues:
                try:
                    d = fdr[pvalue]["Significant"]
                except KeyError:
                    continue

                if d.mObserved == 0:
                    E.info("no data after fdr" )
                    break
                
                elif d.mAverage / d.mObserved < options.fdr_qvalue:
                    E.info("filtering with P-value of %f" % pvalue )
                    if do_filter: data = [ x for x in data if x.mPValue < pvalue ]
                    for d in data:
                        if d.mPValue < pvalue: 
                            d.mFDR = 1
                            d.mQValue = options.fdr_qvalue
                    break
                else:
                    E.warn( "fdr could not be computed - compute more samples (at P = %f, actual fdr=%f)" % (pvalue, d.mAverage / d.mObserved) )
                    no_fdr = True

        if options.fdr_method == "estimate" or (options.fdr_method == "annotator-estimate" and no_fdr):

            E.info( "estimating FDR from observed P-Values" )
            pvalues = [x.mPValue for x in data]
            vlambda = numpy.arange(0, max( pvalues), 0.05 )
            try:
                qvalues = Stats.doFDR( pvalues, vlambda = vlambda, fdr_level = options.fdr_qvalue )
            except ValueError, msg:
                E.warn( "fdr could not be computed - no output: %s" % msg)
                no_fdr = True
            else:
                for d, p, q in zip( data, qvalues.mPassed, qvalues.mQValues ):
                    if p: d.mFDR = 1
                    d.mQValue = q

                if do_filter:
                    data = [ x[0] for x in zip( data, qvalues.mPassed ) if x[1] ]

        if do_filter and no_fdr: data = []

        nremoved = ninput - len(data)

        E.info( "%s: %i data points left, %i removed" % (filename, len(data), nremoved) )

        extra_values = []
        for key in extra_headers:
            if key in input_files:
                extra_values.append( input_files[key] )
            else:
                extra_values.append( "" )

        extra_values = "\t".join( map(str, extra_values ))

        for d in data:
            if d.mFoldChange < 1:
                code = "-"
            else:
                code = "+"

            try:
                id = re.search( options.regex_id, filename ).groups()[0]
            except AttributeError:
                id = filename
            options.stdout.write( "%s\t%s\t%s\t%e\t%6.4f\t%f\t%f\t%f\t%f\t%f\t%i\t%e\t%s\n" % \
                                      (id,
                                       code,
                                       d.mAnnotation,
                                       d.mPValue,
                                       d.mFoldChange,
                                       d.mObserved,
                                       d.mExpected,
                                       d.mCI95[0],
                                       d.mCI95[1],
                                       d.mStdDev,
                                       d.mFDR,
                                       d.mQValue,
                                       extra_values) )

def main():
    parser = E.OptionParser( version = "%prog version: $Id: annotator2tsv.py 2861 2010-02-23 17:36:32Z andreas $", usage = globals()["__doc__"])

    parser.add_option("-g", "--filename-gtf", dest="filename_gtf", type="string",
                      help="filename with gtf information. Used to map a segment to a gene name [default=%default]." )

    parser.add_option("-q", "--query", dest="queries", type="string", action="append",
                      help="term to query with [default=%default]." )

    parser.add_option("-m", "--method", dest="method", type="choice", 
                      choices=("query", "fdr", "fdr-table", "table" ),
                      help="methods to apply [default=%default]." )

    parser.add_option( "--output-format", dest="output_format", type="choice",
                       choices=("table", "gff"),
                       help="output format for query option [default=%default]" )

    parser.add_option( "--fdr-method", dest="fdr_method", type="choice",
                       choices=("annotator", "estimate", "annotator-estimate" ),
                       help="use fdr method. Use annotator, estimate or use annotator and estimate if not enough samples [default=%default].")

    parser.add_option( "--fdr-qvalue", dest="fdr_qvalue", type="float",
                       help="filter output by FDR qvalue (requires annotator output). [default=%default]")

    parser.add_option("-w", "--workspace", dest="filename_workspace", type="string",
                      help="filename with workspace information [default=%default]." )

    parser.add_option("-a", "--annotations", dest="filename_annotations", type="string",
                      help="filename with annotation information [default=%default]." )

    parser.add_option("--annotations-gtf", dest="filename_annotations_gtf", type="string",
                      help="filename with annotation information [default=%default]." )

    parser.add_option("-s", "--segments", dest="filename_segments", type="string",
                      help="filename with segment information [default=%default]." )

    parser.add_option( "--segments-gtf", dest="filename_segments_gtf", type="string",
                      help="filename with segment information [default=%default]." )

    parser.add_option( "--regex-id", dest="regex_id", type="string",
                      help="regular expression to extract id from filename [default=%default]." )

    parser.set_defaults(
        ignore_strand = False,
        filename_workspace = "intronic.workspace",
        filename_annotations = "go_territories.annotations",
        filename_segments = "test.segments",
        filename_segments_gtf = "FastDown.gtf",
        filename_annotations_gtf = "../data/tg1_territories.gff",
        regex_id = "(.*)",
        queries = [],
        method = "fdr",
        fdr_qvalue = None,
        fdr_method = "annotator",
        output_format = "table",
        )

    (options, args) = E.Start( parser )

    if options.method == "query":
        doQuery( options, args )
    elif options.method in ( "fdr", "fdr-table", "table"):
        doFDR( options, args )

    E.Stop()    

if __name__ == "__main__":
    sys.exit(main())




################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
graph_blast2besthits.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python graph_blast2besthits.py --help

Type::

   python graph_blast2besthits.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import getopt
import tempfile
import time
import popen2

USAGE="""python %s [OPTIONS] < graph.in > graph.out

Version: $Id: graph_blast2besthits.py 2782 2009-09-10 11:40:29Z andreas $

Parse blast graph and only output best hits for each query per genome. Note that
the blast graph has been sorted by query_token and sbjct_token.

Options:
-h, --help                      print this message.
-v, --verbose=                  loglevel.
-p, --pattern-genome=           pattern to extract genome
-m, --method=                   method [blast|score|pid]
-s, --score-factor=             thresholding on score (multiplicative)
-i, --pide-factor=              thresholding on pide (additive)
""" % sys.argv[0]

param_long_options = ["help", "verbose=", "pattern-genome=", "method=", "score-factor=", "pide-factor=" ,
                      "version" ]

param_short_options = "hv:p:f:i"

param_loglevel = 1
param_method = "evalue"
param_filename_self_scores = None

import CGAT.Experiment as E
import CGAT.BlastAlignments as BlastAlignments
import math

param_pattern_genome = "^([^|]+)|"
param_method = "evalue"

param_score_threshold_factor = 1.0
param_pide_threshold_factor = 0.0

##-------------------------------------------------------------------------------
def PrintMatches( matches, method=None, filter=None,
                  score_threshold_factor = 1.0,
                  pide_threshold_factor = 0.0):
    """print best match per organism (sorted by E-Value).

    If there are several matches with the same E-Value, all are printed.
    """

    orgs = matches.keys()
    orgs.sort()

    if method == "evalue":
        f1 = lambda x,y: cmp(x.mEvalue, y.mEvalue)
        f2 = lambda x: x.mEvalue
    elif method == "score":
        f1 = lambda x,y:  -cmp(x.score, y.score)
        f2 = lambda x: x.score
    elif method == "pid":
        f1 = lambda x,y:  -cmp(x.mPercentIdentity, y.mPercentIdentity)
        f2 = lambda x: x.mPercentIdentity
    else:
        raise "unknown method"

    noutput = 0
    
    for org in orgs:
        
        if filter and filter == org: continue

        m = matches[org]
        score_threshold = int(m[0].score * score_threshold_factor)
        pide_threshold = int(m[0].mPercentIdentity + pide_threshold_factor)

        if param_loglevel >= 2:
            print "# %i matches for %s, min_score=%i" % (len(m), org, score_threshold)

        if param_loglevel >= 3:
            for mm in m:
                print str(mm)
            
        m.sort( f1 )

        s = f2(m[0])
        x = 0

        if score_threshold_factor != 1.0 and pide_threshold_factor >= 0.0:
            ## take best match and all within score threshold factor
            for x in range(len(m)):
                if m[x].score > score_threshold or \
                       m[x].mPercentIdentity >= pide_threshold:
                    print str(m[x])
                    noutput += 1
        else:
            ## write all with same primary score
            while x < len(m) and s == f2(m[x]) and m[x].score >= score_threshold :            
                print str(m[x])
                x += 1
                noutput +=1
                
    return noutput

##-------------------------------------------------------------------------------
if __name__ == "__main__":

    try:
        optlist, args = getopt.getopt(sys.argv[1:], param_short_options, param_long_options)
    except getopt.error, msg:
        print USAGE, msg
        sys.exit(2)

    for o,a in optlist:
        if o in ( "-v", "--verbose" ):
            param_loglevel = int(a)
        elif o in ( "--version", ):
            print "version="
            sys.exit(0)
        elif o in ( "-h", "--help" ):
            print USAGE
            sys.exit(0)
        elif o in ("-p", "--pattern-genome"):
            param_pattern_genome = a
        elif o in ("-m", "--method"):
            param_method = a
        elif o in ("-s", "--score-factor"):
            param_score_threshold_factor = float(a)
        elif o in ("-i", "--pide-factor"):
            param_pide_threshold_factor = float(a)
            
    print E.GetHeader()
    print E.GetParams()

    ninput, noutput, nskipped, nfailed = 0, 0, 0, 0

    last_query_token = None
    rx = re.compile(param_pattern_genome)
    
    matches = {}
    
    for line in sys.stdin:

        if line[0] == "#": continue

        link = BlastAlignments.Link()
        try:
            link.Read( line )
        except ValueError:
            sys.stderr.write( "parsing error in line %s\n" % line[:-1] )
            continue
        
        ninput += 1
        if link.mQueryToken != last_query_token :
            if last_query_token:
                noutput += PrintMatches( matches,
                                         method = param_method,
                                         filter = rx.search(last_query_token).groups()[0],
                                         score_threshold_factor = param_score_threshold_factor,
                                         pide_threshold_factor = param_pide_threshold_factor )
            matches = {}
            last_query_token = link.mQueryToken

        org = rx.search(link.mSbjctToken).groups()[0]
        
        if org not in matches: matches[org] = []
        matches[org].append( link )

    noutput += PrintMatches( matches,
                             method = param_method,
                             filter = rx.search(last_query_token).groups()[0],
                             score_threshold_factor = param_score_threshold_factor,
                             pide_threshold_factor = param_pide_threshold_factor )
    
    print "# ninput=%i, noutput=%i" % ( ninput, noutput )
    
    print E.GetFooter()

    if noutput == 0:
        if ninput == 0:
            raise "no output, because no input."
        else:
            raise "no output."
        
